#include "Camera.h"

#include <GL/glfw.h>
#include <cmath>

void Camera::init(float x, float y, float z, float yaw, float pitch) {
    place(x, y, z);
    orient(yaw, pitch);
}

void Camera::look() {
    glLoadIdentity();
	gluLookAt(pos.x, pos.y, pos.z,
			  pos.x + dir.x, pos.y + dir.y, pos.z + dir.z,
			  0.0f, 1.0f, 0.0f);
}

void Camera::step(float delta) {
    pos.x += dir.x * delta;
    pos.y += dir.y * delta;
    pos.z += dir.z * delta;
}

void Camera::strafe(float delta) {
    delta /= sqrt(dir.z*dir.z + dir.x*dir.x);
    pos.x -= dir.z * delta;
    pos.z += dir.x * delta;
}

void Camera::pan(float delta) {
	delta /= sqrt(dir.x*dir.x + dir.z*dir.z);
    pos.x += dir.x * delta;
    pos.z += dir.z * delta;
}

void Camera::move(float dx, float dy, float dz) {
    pos.x += dx;
    pos.y += dy;
    pos.z += dz;
}

void Camera::place(float x, float y, float z) {
    pos.x = x;
    pos.y = y;
    pos.z = z;
}

void Camera::turn(float dyaw, float dpitch) {
    orient(yaw + dyaw, pitch + dpitch);
}

void Camera::orient(float nyaw, float npitch) {
    if(npitch < -1.5f)
        npitch = -1.5f;
    else if(npitch > 1.5f)
        npitch = 1.5f;
    
    yaw = nyaw;
    pitch = npitch;

    float r = cos(pitch);
    dir.x = r*sin(yaw);
    dir.y = sin(pitch);
    dir.z = r*cos(yaw);
}
