#include <stdlib.h>
#include <stdio.h>
//#include <math.h>
#include <string.h>
#include "defines.h"
#include "vmath.h"
#include "shader.h"
#include "opengl.h"

#include <cutil.h>
#include <cuda_gl_interop.h>
#include <cutil_inline.h>
#include <cutil_gl_inline.h>

#include "CudaHelpers.h"

extern "C" 
{

void CUDA_Init(int dev) 
{

	int count;
	cudaDeviceProp p;
	
	cudaGetDeviceCount(&count);

	for(int i = 0; i < count; ++i) {
		cudaGetDeviceProperties(&p, i);

		printf("#%d\n", i);
		printf("  Name:				%s\n", p.name );
		printf("  Revision:			%d.%d\n", p.major, p.minor );
		printf("  Global Mem:		%u\n", p.totalGlobalMem );
		printf("  Shared/Blk:		%d\n", p.sharedMemPerBlock );
		printf("  Regs/Blk:			%d\n", p.regsPerBlock );
		printf("  Warp Size:		%d\n", p.warpSize );
		printf("  Mem Pitch:		%d\n", p.memPitch );
		printf("  Thrds/Blk:		%d\n", p.maxThreadsPerBlock );
		printf("  Max block dims:	(%d ,%d,%d)\n", p.maxThreadsDim[0], p.maxThreadsDim[1], p.maxThreadsDim[2]);
		printf("  Max grid dims:	(%d,%d,%d)\n", p.maxGridSize[0], p.maxGridSize[1], p.maxGridSize[2]);
		printf("  Const Mem:		%d\n", p.totalConstMem );
		printf("  Clock Rate:		%d\n", p.clockRate );
		printf("  Can map host memory:		%d\n", p.canMapHostMemory );
	}

	if(dev >= count || dev < 0) {
		dev = cutGetMaxGflopsDeviceId();
	}
    
	printf("CUDA_Init: Using device %d\n", dev);

	CUDA_SAFE_CALL(cudaSetDevice(dev));
	CUDA_SAFE_CALL(cudaGLSetGLDevice(dev));
	
	printf("CUDA_Init: Successfully initialized\n", dev);
}



void CUDA_RegisterVBO(GLuint vbo) 
{
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo));
}

void CUDA_UnregisterVBO(GLuint vbo) 
{
    CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(vbo));
}

void CUDA_CheckError(const char *msg) {
    cudaError_t err = cudaGetLastError();
    if(cudaSuccess != err) {
		printf("Cuda error: %s: %s.\n", msg, cudaGetErrorString(err));	
    }                         
}


void CUDA_AllocateArray(void **devPtr, size_t size)
{
    cutilSafeCall(cudaMalloc(devPtr, size));
}

void CUDA_FreeArray(void *devPtr)
{
    cutilSafeCall(cudaFree(devPtr));
}

void CUDA_ThreadSync()
{
    cutilSafeCall(cudaThreadSynchronize());
}


} // extern "C"

