#ifndef MESH_H_
#define MESH_H_

#include <string>
#include <vector>
#include <map>
#include "opengl.h"
#include "shader.h"
#include "vmath.h"

struct Material
{
    Vec4 ka, kd, ks;
    GLuint texture;
};

struct MaterialCollection
{
    std::vector<std::pair<size_t, std::string> > matStart;
    std::map<std::string, Material> materials;
};

class Mesh
{
private:
    std::vector<Vec3> vertices,
                      normals;
    std::vector<Vec2> texcoords;
    std::vector<std::vector<int> > vIndices,
                                   vnIndices,
                                   vtIndices;

    GLuint ivbo, gvbo;
    GLuint ntex;

    MaterialCollection materials;
public:
    Mesh();
    //Set flipYZ to true to swap y and z while reading the model. 
    bool loadObj(std::string filename, float scale = 1, bool flipYZ = false, float translateY = 0);
    void initBuffers();
    void render();
    Shader shader;
    std::vector<Vec3>& getVertices() { return vertices; }
};

#endif
