#include "Simulator.h"
#include "opengl.h"
#include "glframebufferobject.h"
#include "primitive.h"
#include "shaderprogram.h"
#include "Terrain.h"
#include "gltextureloader.h"
#include "vsml.h"
#include <IL/il.h>
#include <tr1/random>
#include <GL/glfw.h>
#include "Wind.h"
#include "WindLBM.h"
#include "Terrain.h"
#include "Snow.h"
#include "Camera.h"

GLTextureLoader *GLTextureLoader::s_instance = 0;
GLFramebufferManager *GLFramebufferManager::s_instance = 0;
Primitive *pQuad;
Simulator::Simulator(WindowProperties &properties) {
    //renderMode_ = FILL;
    std::cout<<"test"<<std::endl;
    width_ = properties.width;
    height_ = properties.height;

    glClearColor(0.0, 0.0, 0.0, 0.0);
    glViewport(0,0,width_,height_);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glFrontFace(GL_CW);
    glDisable(GL_DITHER);
    
    

    std::cout<<"test"<<std::endl;

    //camera = new Camera(glm::vec2(width_, height_), 60.0f, 1.0f, 7000.0f);
    
    //camera.initCam(glm::vec2(width_, height_), 60.0f, 1.0f, 7000.0f);
    //camera->place(-500.0f, 0.0f, -200.0f);
    //camera->center(0.0, 10.0, 0.0);
    //camera->orient(0, 0);
    
    //camera->rotx = 0.0f;
    //camera->roty = 80.0f;
    
    camera_.center = glm::vec3(0.0, 10.0, 0.0);
    camera_.eye = glm::vec3(200.0, 100.0, 100.0);
    camera_.up = glm::vec3(0.0, 1.0, 0.0);
    camera_.near = 2.0;
    camera_.far = 1000000.0f;
    camera_.rotx = 0.f; camera_.roty = 80.f;
    camera_.fovy = 45.0;
    
    GLFramebufferObjectParams params;
    params.width = properties.width;
    params.height = properties.height;
    params.hasDepth = true;
    params.depthFormat = GL_DEPTH_COMPONENT16;
    params.format = GL_RGBA16F;
    params.nColorAttachments = 2;
    params.nSamples = 32;//GLFramebufferObject::sssdqueryMaxSamples();
    params.nCSamples = 32;
    params.type = GL_TEXTURE_2D;
    (*GLFramebufferManager::instance()->framebuffers())["m"] = new GLFramebufferObject(params);
    
    params.hasDepth = true;
    params.nSamples = 0;

    (*GLFramebufferManager::instance()->framebuffers())["0"] = new GLFramebufferObject(params);
    params.hasDepth = true;
    params.nColorAttachments = 1;
    (*GLFramebufferManager::instance()->framebuffers())["1"] = new GLFramebufferObject(params);
    
    primitives_["quad1"] = new Quad(glm::vec3(1, 1, 0), glm::vec3(width_ * 0.5, height_ * 0.5, 0), glm::vec3(width_, height_, 1));

    primitives_["plane0"] = new Plane(glm::vec3(10, 0, 10), glm::vec3(0, 0, 0), glm::vec3(20, 1, 20));
    
    
    
    TerrainParams paramsT;
    paramsT.resolution = 128;
    paramsT.gain = 0.61;
    paramsT.grid = glm::vec2(24,24);
    paramsT.lacunarity = 1.7;
    paramsT.offset = 1;
    paramsT.noiseScale = .15;
    paramsT.tess = glm::vec3(512, 512, 512);
    paramsT.octaves = 12;
    paramsT.random = glm::vec2(0.f, 0.f);

    //terrain_ = new TessellationTerrain(paramsT, this);
    lightPos_ = glm::normalize(glm::vec3(1.0, 0.1, 0.0)); 
}


Simulator::~Simulator() {
    
    for (auto itr = shaderPrograms_.begin(); itr != shaderPrograms_.end(); ++itr)
	 delete (*itr).second;
    for (auto itr = primitives_.begin(); itr != primitives_.end(); ++itr)
	 delete (*itr).second;

    delete terrain_;
    delete snow;
    if(wind)
        delete wind;
    if(windLBM)
        delete windLBM;
        
    delete terrain;
    
}

void Simulator::resize(int w, int h) {
    width_ = w; height_ = h;
    glViewport(0, 0, width_, height_);
    primitives_["quad1"]->tesselate(glm::vec3(1, 1, 0), glm::vec3(width_ * 0.5, height_ * 0.5, 0), glm::vec3(width_, height_, 1));
 
    (*GLFramebufferManager::instance()->framebuffers())["m"]->resize(width_, height_);
    (*GLFramebufferManager::instance()->framebuffers())["0"]->resize(width_, height_);
    (*GLFramebufferManager::instance()->framebuffers())["1"]->resize(width_, height_);
    
    maxMipLevel_ = (log10((float)max(width_, height_)) / log10(2.f));
   
}

void Simulator::init()
{
    snow = new Snow();
    wind = new Wind();
    windLBM = new WindLBM();
    terrain = new Terrain();

    vector<Mesh*> meshes;
    Road road(conf.roadDefinitionFile, SCENE_X/float(conf.map.width*conf.map.resX), conf.map.minHeight);
    meshes.push_back(&(road.getMesh()));
    
    terrain->Init(conf.map.heightmap, conf.map.normalmap, conf.map.width, conf.map.height, road);
    std::cout << "Terrain loaded (with " << conf.map.width << "x" << conf.map.height << " heightmap)" << "\n";
    std::cout << "Heightmap loaded from " << conf.map.heightmap << ", normalmap from " << conf.map.normalmap << "\n\n";

    std::cout << "Wind field initialization started";
    if(conf.usingLBM)
        std::cout << " (LBM simulation)";
    else
        std::cout << " (finite difference Navier-Stokes)";
    std::cout << "\n\n";

    if(conf.usingLBM)
        windLBM->Init(conf.windX, conf.windY, conf.windZ, conf.map.width, conf.LBMOmega, conf.LBMViscosity, conf.LBMVelFactor, conf.LBMInitVel, conf.LBMInitRho, conf.LBMBoundaryVelFactor);
    else
        wind->Init(conf.windX, conf.windY, conf.windZ, conf.map.width);

    snow->Init(conf.numParticles);
    
    std::cout << "Updating obstacles\n";
    int layers[] = { 1, 1, conf.windY-1, 1, conf.windY-2, 1, 1 };
    for(int i = 0; i < 7; ++i) {
        for(int y = 0; y < layers[i]; ++y)
            snow->UpdateObstacles(terrain->gvbo, i, y);
    }
    
}

void Simulator::UpdateObstacles(int stage, int obs_y)
{
    snow->UpdateObstacles(terrain->gvbo, stage, obs_y);
}



static bool stepMode = false;
static int stepCount = 0;
static float windLimit = 60.0f;
static std::vector<float> mlupsEntries;
static float temp_mlups = 0;

void Simulator::update(float dt)
{

    if(conf.simulateWind) {
            if(conf.usingLBM)
            {
                windLBM->SimulateWindField(conf.LBMCollideKernel, 0.05f, stepMode, stepCount, conf.enableDisplay, temp_mlups);
            }
            else
            {
                wind->SimulateWindField(0.05f, stepMode, stepCount, conf.enableDisplay);
            }
        }

        if(conf.simulateSnow)
            snow->MoveParticles(snow->vbo, terrain->gvbo, snow->count, 0.05f);

}
static bool firstTime = true;
void Simulator::draw(Camera& camera) {
    setPersepectiveProjection();
    
    glEnable(GL_DEPTH_TEST);
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
    
    //terrain_->draw();
     
    terrain->Render();
    
    if(conf.renderSnow) {
        snow->Render(&camera_);
    }
    
    
}

void Simulator::drawStereo() {
    
}

void Simulator::setOrthoProjection() {
    auto vsml = VSML::getInstance();
    vsml->loadIdentity(VSML::PROJECTION);
    vsml->ortho(0.f,(float)width_,(float)height_,0.f);
    vsml->loadIdentity(VSML::MODELVIEW);
}

void Simulator::setOrthoProjection(int width, int height) {
    auto vsml = VSML::getInstance();
    vsml->loadIdentity(VSML::PROJECTION);
    vsml->ortho(0.f,(float)width,(float)height,0.f);
    vsml->loadIdentity(VSML::MODELVIEW);
}

void Simulator::setPersepectiveProjection() {
    auto vsml = VSML::getInstance();
    int w = width_, h = height_;
    vsml->perspective(camera_.fovy, w / (float)h, camera_.near, camera_.far);
    vsml->loadIdentity(VSML::MODELVIEW);
     
    vsml->rotate(camera_.rotx, 1.f, 0.f, 0.f);
    vsml->rotate(camera_.roty, 0.f, 1.f, 0.f);
    vsml->translate(-camera_.eye.x, -camera_.eye.y, -camera_.eye.z);
}

float sensitivity = 0.1f;
void Simulator::mouseMove(float dx, float dy) {

    auto dxs = dx*sensitivity;
    auto dys = dy*sensitivity;  
    camera_.rotx += dys;
    camera_.roty -= dxs; 
}
std::tr1::mt19937 prng(1337);
std::tr1::uniform_real<float> uniform;
std::tr1::variate_generator<std::tr1::mt19937, std::tr1::uniform_real<float> > randu(prng,uniform);
void Simulator::processKeyEvents(const int key) {

    switch(key) {
            case '1':
                //this->setRenderMode(FILL);
                break;
            case '2':
                //this->setRenderMode(WIREFRAME);
                break;
            case '.':
                terrain_->setLod(terrain_->lod() + 1);
                break;
            case ',':
                terrain_->setLod(max(terrain_->lod() - 1, 1.f));
                break;
            case 'G':
                {
                    TerrainParams paramsT = terrain_->params();
	                delete terrain_;
	                terrain_ = 0;
	                paramsT.random = glm::vec2(randu()*100000.f, randu()*100000.f);
	                terrain_ = new TessellationTerrain(paramsT, this);
	                break;
                }
 
        }
}

void Simulator::step(const float delta)
{ 
    float yrotrad = camera_.roty / 180 * 3.141592654f;
	float xrotrad = camera_.rotx / 180 * 3.141592654f;
	camera_.eye.x += sinf(yrotrad)*delta;
	camera_.eye.z -= cosf(yrotrad)*delta;
	camera_.eye.y -= sinf(xrotrad)*delta;
}

void Simulator::strafe(const float delta)
{
    float yrotrad = (camera_.roty / 180 * 3.141592654f);
	camera_.eye.x += cosf(yrotrad)*delta;
	camera_.eye.z += sinf(yrotrad)*delta;
}

void Simulator::move(const float x, const float y, const float z)
{
    camera_.eye.y += y;
}

