#ifndef SIMULATOR_H
#define SIMULATOR_H

#include <unordered_map>
#include <cstring>
#include "opengl.h"
#include "glm/glm.hpp"
#include "config.h"


struct Camera_ {
    glm::vec3 eye, center, up;
    float fovy, near, far;
    float rotx, roty;
    void perspective_camera(int w,int h) {

	float ratio = w / static_cast<float>(h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fovy,ratio,near,far);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(rotx,1.0,0.0,0.0);  //rotate our camera on teh

	glRotatef(roty,0.0,1.0,0.0);
	glTranslatef(-eye.x, -eye.y, -eye.z);
	//glLoadIdentity();
    }


    void orthogonal_camera(int w,int h) {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,static_cast<float>(w),static_cast<float>(h),0.f,-1.f,1.f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    }
};

class Primitive;
class ShaderProgram;
class VSML;
class TessellationTerrain;
class Snow;
class Wind;
class WindLBM;
class Terrain;
class Camera;

struct WindowProperties {
    int width, height;
};

class Simulator
{
public:
    Simulator(WindowProperties &properties);
    ~Simulator();

    void init();

    void resize(int w, int h);
    void draw(Camera &cam);
    void drawStereo();
    void update(float dt);
    void mouseMove(float dx, float dy);

    int width() { return width_; }
    int height() { return height_; }
    void setPersepectiveProjection();
    void setOrthoProjection();
    void setOrthoProjection(int width, int height);
    
    Camera_ *getCamera() { return &camera_; }
    glm::vec3 light() { return lightPos_; }
    
    void processKeyEvents(const int key);
    void step(const float delta);
    void strafe(const float delta);
    void move(const float, const float, const float);
    
    void UpdateObstacles(int, int);
protected:
    int width_, height_;
    Camera_ camera_;
    glm::vec3 lightPos_;
    std::unordered_map<const char*, ShaderProgram *> shaderPrograms_;
    std::unordered_map<const char*, Primitive *> primitives_;

    int maxMipLevel_;
    TessellationTerrain *terrain_;
    Terrain *terrain;
    Snow *snow;
    Wind *wind;
    WindLBM *windLBM;
};
#endif
