#include "Snow.h"
#include "soil/SOIL.h"
#include "vmath.h"
#include <cstdlib>
#include <fstream>
#include <algorithm>
#include "defines.h"
#include "utils.h"

#include "config.h"
#include "Camera.h"

#include "CudaHelpers.h"

#include "ParticleSystem.h"

#include <iostream>

#include "vsml.h"
#include "shaderprogram.h"

struct VecCmp {
    bool operator()(const Vec4 &a, const Vec4 &b) const {
	    int ax, ay, az;
	    int bx, by, bz;

	    ax = (int)a.x;
	    ay = (int)a.y;
	    az = (int)a.z;

	    bx = (int)b.x;
	    by = (int)b.y;
	    bz = (int)b.z;

	    if((ay & ~3) < (by & ~3))
		    return true;
	    if((ax & ~15) < (bx & ~15))
		    return true;
	    if((az & ~15) < (bz & ~15))
		    return true;
	    /*if((ax & ~15) < (bx & ~15))
		    return true;
	    if((az & ~15) < (bz & ~15))
		    return true;*/
	    return false;

	    /*if(a.y < b.y)
		    return true;
	    if(a.z < b.z)
		    return true;
	    if(a.x < b.x)
		    return true;
	    return false;*/
    }
};


static float calc_theta() {
	return float(myrandom() * 2 * M_PI);
}

Snow::~Snow() {
        if(hasAllocated) {        
                printf("Running snow desctructor.\n");
	        CUDA_Snow_FreeParticles();
	        glDeleteBuffers(1, &vbo);
                printf("Snow destructor done.\n");
        }
}

Snow::Snow()
{
        hasAllocated = false;
}

/**
 * Snow particle simulation constructor
 * @param count number of snow particles to simulate
 * @param altPartInit true if the alternate particle initialization method is to be used
 */
void Snow::Init(GLsizei count) {
	this->count = count;

	std::cout << "Snow::Init()" << std::endl;

	tex = SOIL_load_OGL_texture(
		//"particle.png",
		"particle2.png",
		SOIL_LOAD_AUTO,
		SOIL_CREATE_NEW_ID,
		SOIL_FLAG_MIPMAPS
	);

	//shader.load("snow.vert", "snow.frag");

	Vec4 *data = new Vec4[count];
	for(int i = 0; i < count; ++i) {
		data[i] = Vec4(
			myrandom() * SCENE_X,
			myrandom() * SCENE_Y,
			myrandom() * SCENE_Z,
			calc_theta()
		);
//                std::cout << "SCENE_Y: " << SCENE_Y << " Y-position: " << data[i].y << std::endl;
	}

	//std::sort(data, data+count, VecCmp());

	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	GLsizei size = count * 4 * sizeof(float);
	glBufferData(GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	glGenVertexArrays(1, &vba);
	glBindVertexArray(vba);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, NULL);
	glEnableVertexAttribArray(0);
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);

	CUDA_RegisterVBO(vbo);

        CUDA_Snow_InitParticles(count, conf.gravity, conf.snowGrowth);

	delete [] data;

        hasAllocated = true;
        
    snowShader = new ShaderProgram();
    snowShader->loadShaderFromSource(GL_VERTEX_SHADER, "shaders/snow.glsl");
    snowShader->loadShaderFromSource(GL_FRAGMENT_SHADER, "shaders/snow.glsl");
    snowShader->link();
}

void Snow::MoveParticles(GLuint vbo, GLuint ovbo, int length, float deltaTime)
{
//		std::cout<<"before move particle"<<std::endl;
        CUDA_Snow_MoveParticles(vbo, ovbo, length, deltaTime);
//		std::cout<<"after move particle"<<std::endl;
		//std::cout<<(cudaGetErrorString(cudaGetLastError()))<<std::endl;
		//std::cin.get();
}

void Snow::UpdateObstacles(unsigned int ovbo, int stage, int y)
{

        CUDA_WindField_UpdateObstacles(ovbo, stage, y);

}

bool Snow::Sort() {
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	Vec4 *data = (Vec4 *)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_ONLY);

	if(!data)
		return false;

	//std::sort(data, data+count, vec_cmp);

	glUnmapBuffer(GL_ARRAY_BUFFER);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	return true;
}

void Snow::Render(Camera_ *cam) {
	/*glDepthMask(GL_FALSE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);


	glColor4f(1.0f, 1.0f, 1.0f, 0.3f);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glVertexPointer(4, GL_FLOAT, 0, 0);
	glDrawArrays(GL_POINTS, 0, count);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	
	glDisable(GL_BLEND);
	glDepthMask(GL_TRUE);
	*/

	glPointSize(10.0f);

	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex);

	//shader.bind();
	//shader.uniform("colormap", 0);
	//shader.uniform("type", 0);
	//shader.uniform("eye", cam->pos.x, cam->pos.y, cam->pos.z);

    auto vsml = VSML::getInstance();
    snowShader->bind();
    snowShader->setUniformValue("colormap", 0);
	snowShader->setUniformValue("type", 0);
	snowShader->setUniformValue("eye", cam->eye.x, cam->eye.y, cam->eye.z);
	snowShader->setUniformValue("projMatrix2", vsml->get(VSML::PROJECTION, false));
	snowShader->setUniformValue("modelviewMatrix2", vsml->get(VSML::MODELVIEW, false));
    	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_POINT_SPRITE);
	glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);
	
	glDepthMask(GL_FALSE);
	

	glBindVertexArray(vba);
	glDrawArrays(GL_POINTS, 0, count);
	glBindVertexArray(0);
	
	/*
#if 1
	shader.uniform("type", 1);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glVertexPointer(4, GL_FLOAT, 0, 0);
	glDrawArrays(GL_POINTS, 0, count/32);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
#endif
*/
	glDisable(GL_POINT_SPRITE);
	glDisable(GL_BLEND);
	glDepthMask(GL_TRUE);

    snowShader->release();
	//shader.unbind();
	glDisable(GL_TEXTURE_2D);
}
