#include "defines.h"
#include "utils.h"
//#include "WindSystemKernels.cuh"
//#include <cutil.h>



//#ifdef FERMI_OPT
//	#define __mul24(A,B) A*B
//#else
//	#define __mul24(A,B) __mul24(A,B)
//#endif

__constant__ float gravity;
__constant__ float snow_growth;
__constant__ float delta_time;

__device__ void incrementHeight(int x, int y, float amount, float4 *grid) {
	if(x < 0) x = 0;
	if(y < 0) y = 0;
	if(x >= terrain_dim) x = terrain_dim-1;
	if(y >= terrain_dim) y = terrain_dim-1;

	//int ix = __mul24((terrain_dim, y) + x;
	int ix = terrain_dim * y + x;
	//if(grid[ix].w < 1.0f)
		grid[ix].w += amount*snow_growth;
}

__device__ int2 calcGridPos(float4 p) {
    int2 gridPos;
    gridPos.x = floor((float)terrain_dim*p.x/(float)SCENE_X);
    gridPos.y = floor((float)terrain_dim*p.z/(float)SCENE_Z);

    if(gridPos.x < 0) gridPos.x = 0;
    if(gridPos.y < 0) gridPos.y = 0;
    if(gridPos.x >= terrain_dim) gridPos.x = terrain_dim-1;
    if(gridPos.y >= terrain_dim) gridPos.y = terrain_dim-1;

    return gridPos;
}

#define SG_DIM 16
#define SG_STRIDE 18

__constant__ int sg_off[4] = { -1, -SG_STRIDE, 1, SG_STRIDE };

__global__ void smooth_ground(float4 *grid) {
	const int tx = threadIdx.x;
	const int ty = threadIdx.y;

	volatile int _0 = 0;
	volatile int _1 = 1;

	float4 *block = (float4 *)shared + (ty+_1)*SG_STRIDE + tx+_1;
	grid += (blockIdx.y*blockDim.y + ty)*terrain_dim + blockIdx.x*blockDim.x + tx;

	block[0] = grid[0];

	if(tx == _0) {
		if(blockIdx.x == _0)
			block[-_1] = block[0];
		else
			block[-_1] = grid[-_1];
	}
	if(tx == blockDim.x-_1) {
		if(blockIdx.x == gridDim.x-_1)
			block[_1] = block[0];
		else
			block[_1] = grid[_1];
	}
	if(ty == _0) {
		if(blockIdx.y == _0)
			block[-SG_STRIDE] = block[0];
		else
			block[-SG_STRIDE] = grid[-terrain_dim];
	}
	if(ty == blockDim.y-_1) {
		if(blockIdx.y == gridDim.y-_1)
			block[SG_STRIDE] = block[0];
		else
			block[SG_STRIDE] = grid[terrain_dim];
	}

	__syncthreads();

	float mod = 0.0f;
	const float a = block[0].y + block[0].w;
#pragma unroll
	for(int i = _0; i < 4; ++i) {
		int k = sg_off[i];
		float b = block[k].y + block[k].w;
		float diff = b - a;

		if(fabs(diff) > 0.05f) {
			// I'm highest
			if(diff < _0) {
				if(block[0].w > 0.1f) {
					float snow = block[0].w;
					if(-diff < snow)
						snow = -diff;
					mod -= 0.05f * snow;
				}
			}
			else {
				if(block[k].w > 0.1f) {
					float snow = block[k].w;
					if(diff < snow)
						snow = diff;
					mod += 0.05f * snow;
				}
			}
		}
	}

	__syncthreads();

	block[0].w += mod;
	grid[0] = block[0];
}

__device__ void reposition(float4 &pos) {
	int or_term = 0x3F000000 | ((*(int*)&pos.y & 0xFF) << 2);
	float temp = 4.0f*SCENE_X;

	int ival = ((*(int*)&pos.x & 0xFFF) << 10) | or_term;
	pos.x = (*(float*)&ival - 0.5f) * temp;
	
	ival = ((*(int*)&pos.z & 0xFFF) << 10) | or_term;
	pos.z = (*(float*)&ival - 0.5f) * temp;

	pos.y = SCENE_Y - 2.0f;
}

__global__ void part_update(float4 *part_pos, float4 *part_vel, float4 *grid) {
	//int tid = __mul24((blockIdx.x, blockDim.x) + threadIdx.x;
	int tid = blockIdx.x * blockDim.x + threadIdx.x;

	float4 pos = part_pos[tid];
	float4 vel = part_vel[tid];
//	float4 v_drag = tex3D(wind_vel_tex, pos.x + 1.0f, pos.z + 1.0f, pos.y+ 1.0f);	
	float4 v_drag = tex3D(wind_vel_tex, pos.x*convert[0] + 1.0f, pos.z*convert[2] + 1.0f, pos.y*convert[1] + 1.0f);	
	
//	if(tid > 10 && tid < 30) {
//		printf("Pos: %f, %f, %f\n", pos.x, pos.y, pos.z);
//		printf("Particle vel: %f, %f, %f\n", vel.x, vel.y, vel.z);
//		printf("Drag vel: %f, %f, %f\n", v_drag.x, v_drag.y, v_drag.z);
//	}

//    v_drag.x = v_drag.y = v_drag.z = 1;
//    v_drag.x = 1;
	v_drag.x -= vel.x;
	v_drag.y -= vel.y;
	v_drag.z -= vel.z;
	float temp = v_drag.x*v_drag.x + v_drag.y*v_drag.y + v_drag.z*v_drag.z;
	float v_fluid_abs = sqrtf(temp);
//	temp *= 1.0f / v_fluid_abs;
    temp = v_fluid_abs;
	v_drag.x *= temp*gravity*vel.w;
	v_drag.y *= temp*gravity*vel.w;
	v_drag.z *= temp*gravity*vel.w;
//	v_drag.x *= gravity*vel.w;
//	v_drag.y *= gravity*vel.w;
//	v_drag.z *= gravity*vel.w;

    float ax = v_drag.x;
    float ay = -gravity + v_drag.y;
    float az = v_drag.z;

//	float4 vel_old = make_float4(vel.x, vel.y, vel.z, vel.w);

	float rot_omega = omegas[tid & 31];
	pos.w += rot_omega*delta_time;
	rot_omega *= radiuses[tid & 31];
	float v_abs = max(sqrtf(vel.x*vel.x + vel.y*vel.y + vel.z*vel.z), 1.0f);
	rot_omega *= v_fluid_abs / v_abs;
	float v_circ_x = -__sinf(pos.w) * rot_omega;
	float v_circ_z = __cosf(pos.w) * rot_omega;
//	v_circ_x = 0;
//	v_circ_z = 0;

	pos.x += (vel.x + v_circ_x)*delta_time + 0.5f*ax*delta_time*delta_time;
	pos.y += vel.y*delta_time + 0.5f*ay*delta_time*delta_time;
	pos.z += (vel.z + v_circ_z)*delta_time + 0.5f*az*delta_time*delta_time;

    vel.x += ax*delta_time;
	vel.y += ay*delta_time;
	vel.z += az*delta_time;
	part_vel[tid] = vel;

	if(pos.y < -1.0f) pos.y = SCENE_Y-2.0f;//+= SCENE_Y;
	else if(pos.y > SCENE_Y) {
		reposition(pos);
		float4 wind = tex3D(wind_vel_tex, pos.x*convert[0] + 1.0f, pos.z*convert[2] + 1.0f, pos.y*convert[1] + 1.0f);
		part_vel[tid] =  make_float4(wind.x, wind.y, wind.z, vel.w);
	}
	
    int moved = 0;

    if(pos.x < 0.0f) { 
        pos.x += SCENE_X;
//        pos.x = (int)pos.x % (int)SCENE_X + SCENE_X; moved = 1; 
//        pos.x = 0;
        moved = 1;
    }
    else if(pos.x > SCENE_X) { 
        pos.x -= SCENE_X;
//        pos.x = (int)pos.x % (int)SCENE_X; moved = 1; 
//        pos.x = 0;
        moved = 1;
    }
    if(pos.z < 0.0f) { 
        pos.z += SCENE_Z;
//        pos.z = (int)pos.z % (int)SCENE_Z + SCENE_Z; moved = 1; 
//        pos.z = 0;
        moved = 1;
    }
    else if(pos.z > SCENE_Z) { 
        pos.z -= SCENE_Z;
//        pos.z = (int)pos.z % (int)SCENE_Z; moved = 1; 
//        pos.z = 0;
        moved = 1;
    }

#if 1
	/*if(pos.y < TERRAIN_MAX)*/  {
		int2 ip = calcGridPos(pos);

		//float4 hv = make_float4(0, 0, 0, 0);
		//float4 hv = grid[__mul24((terrain_dim, ip.y) + ip.x];
		float4 hv = grid[terrain_dim * ip.y + ip.x];
		float h = hv.y + hv.w;

		if(pos.y < h) {
			if(moved) {
				reposition(pos);
				float4 wind = tex3D(wind_vel_tex, pos.x*convert[0] + 1.0f, pos.z*convert[2] + 1.0f, pos.y*convert[1] + 1.0f);
				part_vel[tid] =  make_float4(wind.x, wind.y, wind.z, vel.w);
			}
			else if (ip.x != terrain_dim-1 && ip.y != terrain_dim-1) {
				incrementHeight(ip.x, ip.y, 0.15f, grid);
				incrementHeight(ip.x+1, ip.y, 0.095f, grid);
				incrementHeight(ip.x-1, ip.y, 0.095f, grid);
				incrementHeight(ip.x, ip.y+1, 0.095f, grid);
				incrementHeight(ip.x, ip.y-1, 0.095f, grid);
				incrementHeight(ip.x+1, ip.y+1, 0.059f, grid);
				incrementHeight(ip.x+1, ip.y-1, 0.059f, grid);
				incrementHeight(ip.x-1, ip.y+1, 0.059f, grid);
				incrementHeight(ip.x-1, ip.y-1, 0.059f, grid);

				incrementHeight(ip.x+2, ip.y, 0.026f, grid);
				incrementHeight(ip.x-2, ip.y, 0.026f, grid);
				incrementHeight(ip.x, ip.y+2, 0.026f, grid);
				incrementHeight(ip.x, ip.y-2, 0.026f, grid);

				incrementHeight(ip.x+2, ip.y+1, 0.015, grid);
				incrementHeight(ip.x+2, ip.y-1, 0.015, grid);
				incrementHeight(ip.x-2, ip.y+1, 0.015, grid);
				incrementHeight(ip.x-2, ip.y-1, 0.015, grid);
				incrementHeight(ip.x+1, ip.y+2, 0.015, grid);
				incrementHeight(ip.x+1, ip.y-2, 0.015, grid);
				incrementHeight(ip.x-1, ip.y+2, 0.015, grid);
				incrementHeight(ip.x-1, ip.y-2, 0.015, grid);

				incrementHeight(ip.x+2, ip.y+2, 0.0037, grid);
				incrementHeight(ip.x+2, ip.y-2, 0.0037, grid);
				incrementHeight(ip.x-2, ip.y+2, 0.0037, grid);
				incrementHeight(ip.x-2, ip.y-2, 0.0037, grid);
//				incrementHeight(ip.x, ip.y, 1.0f, grid);
//				incrementHeight(ip.x+1, ip.y, 0.5f, grid);
//				incrementHeight(ip.x-1, ip.y, 0.5f, grid);
//				incrementHeight(ip.x, ip.y+1, 0.5f, grid);
//				incrementHeight(ip.x, ip.y-1, 0.5f, grid);
//				incrementHeight(ip.x+1, ip.y+1, 0.25f, grid);
//				incrementHeight(ip.x+1, ip.y-1, 0.25f, grid);
//				incrementHeight(ip.x-1, ip.y+1, 0.25f, grid);
//				incrementHeight(ip.x-1, ip.y-1, 0.25f, grid);
				
				reposition(pos);
				float4 wind = tex3D(wind_vel_tex, pos.x*convert[0] + 1.0f, pos.z*convert[2] + 1.0f, pos.y*convert[1] + 1.0f);
//                if (wind.y == 0) wind.y = -0.2;
				part_vel[tid] =  make_float4(wind.x, wind.y, wind.z, vel.w);
			}
		}
	}
#endif

	part_pos[tid] = pos;
}

