#include <fstream>
#include <iostream>
#include <cstdlib>
#include <vector>
#include <algorithm>
#include <limits>
#include <map>
#include "Terrain.h"
#include "vmath.h"
#include "soil/SOIL.h"
#include "config.h"
#include "CudaHelpers.h"
//NEW INCLUDES
#include "Simulator.h"
#include "primitive.h"
#include "glframebufferobject.h"
#include "gltextureloader.h"
#include <IL/il.h>
#include "vsml.h"


using namespace std;
using namespace clothoid;

static float normalize_y(ushort val) {
	float y = (float)val;
	y /= 65536.0f;
	return y;
}

double getPointNearest(double x, double y, int width, int height, Vec4* geo) 
{
	int i = round(y),
	    j = round(x);

	i = std::max(std::min(i, height-1), 0);
	j = std::max(std::min(j, width-1), 0);

	return geo[i*width+j].y;
}

double getPointBilinear(double x, double y, int width, int height, Vec4* geo) 
{
	double y_1 = floor(y);
	double y_2 = ceil(y);
	if (y_2 == y_1) y_2 += 1;

	double x_1 = floor(x);
	double x_2 = ceil(x);
	if (x_1 == x_2) x_2 += 1;

	double fx_1 = getPointNearest(x_1, y_1, width, height, geo)*(x_2-x)/(x_2-x_1) + getPointNearest(x_2, y_1, width, height, geo)*(x-x_1)/(x_2-x_1);
	double fx_2 = getPointNearest(x_1, y_2, width, height, geo)*(x_2-x)/(x_2-x_1) + getPointNearest(x_2, y_2, width, height, geo)*(x-x_1)/(x_2-x_1);
	double fp = fx_1*(y_2-y)/(y_2-y_1) + fx_2*(y-y_1)/(y_2-y_1);

	return fp;
}


Terrain::~Terrain() {
	printf("Running terrain destructor.\n");
	glDeleteBuffers(1, &gvbo);
	glDeleteBuffers(1, &ivbo);
	glDeleteTextures(1, &ntex);
	glDeleteTextures(1, &snowtex);
	glDeleteTextures(1, &groundtex);
	delete[] geo;
	printf("Terrain destructor done.\n");
}

static Primitive* quad_;
void Terrain::Init(const std::string &hname, const std::string &nname, const uint width, const uint height, Road& road) {
glCheckError("Failed to start creating normal texturesdfsdfaa");
    quad_ = new Plane(glm::vec3(width, 1, height), glm::vec3(width, 0, height), glm::vec3(SCENE_X, 1, SCENE_Z));
	const uint size = width * height;
	const uint nsize = 3 * (width) * (height);
	const uint tsize = 4 * (width) * (height);

	// Heightmap read from file
	ushort *hmap = new ushort[size];
	// Normalmap read from file
	uchar *nmap = new uchar[nsize];
	if (geo) delete[] geo;
	// 3D array of the terrain coordinates
	geo = new Vec4[size];

	// Open the heightmap file
	std::ifstream hfile(hname.c_str(), std::ios::binary);
	hfile.read((char *)hmap, size*sizeof(ushort));
	
	float *tmap = new float[tsize];
	
	

	double height_scaling = SCENE_X / ((double)width*conf.map.resX) * (conf.map.maxHeight - conf.map.minHeight);

	// Loop through the XY-plane
	for(uint y = 0; y < height; ++y) {
		for(uint x = 0; x < width; ++x) {
			Vec4 &v = geo[y*width+x];
			v.x = SCENE_X*(float)x/(float)width;
			
			v.y = normalize_y(hmap[y*width+x]) * height_scaling;
            tmap[(y*width+x)*4] = v.y;
            tmap[(y*width+x)*4+1] = 0;
            tmap[(y*width+x)*4+2] = 0;
            tmap[(y*width+x)*4+3] = 0;
			//             Uncomment to get a flat terrain
			            v.y = 1.0;

			v.z = SCENE_Z*(float)y/(float)height;

			// Create a block on the ground for wind interaction

			//            if(v.x > 50 && v.x < 70 && v.z > 50 && v.z < 70) {
			//                v.y = 30;
			//            }
			/*
			   v.y = normalize_y(4000);
			   int w = 16;
			   if(x > dim/2-w && y > dim/2-w && x < dim/2+w && y < dim/2+w)
			   v.y = normalize_y(40000);
			 */

			v.w = 0;
		}
	}
	
	if (ttex)
		glDeleteTextures(1, &ttex);

	// Bind normalmap to a texture
	glGenTextures(1, &ttex);
	glBindTexture(GL_TEXTURE_2D, ttex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, width, height, 0, GL_RGBA32F, GL_FLOAT, tmap);
/*

	// Adjust terrain to road
	// Generate distance map (first parameter is distance, second is elevation)
	pair<float,float> *dmap = new pair<float,float>[size];
	for (size_t i = 0; i < size; i++)
		dmap[i] = pair<float,float>(numeric_limits<float>::infinity(), 0);

	//Go through the vertices of the road mesh and update the distance map
	vector<Vec3>& verts = road.getMesh().getVertices();
	printf("Adjusting terrain...\n"); fflush(stdout);
	for (size_t i = 0; i < verts.size(); i++)
	{
		Vec3 p = verts[i];
		p.x *= (float)width/SCENE_X;
		p.z *= (float)width/SCENE_X;
		//find nearest point
		int x = round(p.x),
		    y = round(p.z);
		double elev = p.y;

		for (int k1 = -1; k1 <= 1; k1++)
		{
			if (x+k1 < 0 || x+k1 >= (int)width) continue;

			for (int k2 = -1; k2 <= 1; k2++)
			{
				if (y+k2 < 0 || y+k2 >= (int)height) continue;

				Vec2i pp(x+k1, y+k2);
				double dx = abs(pp.x - p.x),
				       dy = abs(pp.y - p.z);
				double dist = length(Vec2d(dx,dy));
				if (dmap[pp.y*width+pp.x].first > dist)
				{
					dmap[pp.y*width+pp.x].first = dist;
					dmap[pp.y*width+pp.x].second = elev;
				}
			}
		}
	}
	*/
/*
	for (size_t i = 0; i < height; i++)
	{
		for (size_t j = 0; j < width; j++)
		{
			if (dmap[i*width+j].first < width*height*conf.map.resX*conf.map.resY)
				geo[i*width+j].y = dmap[i*width+j].second-0.1;
		}
	}
	*/
        //delete[] dmap;
	printf("Done adjusting.\n"); fflush(stdout);

	printf("Generating normal map...\n"); fflush(stdout);
	for (size_t i = 0; i < height; i++)
	{
		for (size_t j = 0; j < width; j++) 
		{
			float sj = geo[i*width+(j<width-1?j+1:j)].y - geo[i*width+(j>0?j-1:j)].y,
			      si = geo[(i<height-1?i+1:i)*width+j].y - geo[(i>0?i-1:i)*width+j].y;

			if (i == 0 || i == height-1) si *= 2;
			if (j == 0 || j == width-1) sj *= 2;

			Vec3 n = normalize(Vec3(-sj, 2., si));
			//            Vec3 n = normalize(Vec3(-sj*height_scaling, 2.*SCENE_X/height_scaling, si*height_scaling));
			//            Vec3 n = normalize(Vec3(-sj*height_scaling, 2.*SCENE_X/((float)width*conf.map.resX), sj*height_scaling));
			nmap[(i*width+j)*3] = (int)(max(min(n.x*128.+128., 255.),0.));
			nmap[(i*width+j)*3+1] = (int)(max(min(n.y*128.+128., 255.),0.));
			nmap[(i*width+j)*3+2] = (int)(max(min(n.z*128.+128., 255.),0.));

			continue;

			Vec4 e1 = geo[i*width+j],
			     e2 = geo[(i < width-1 ? (i+1) : i)*width+j],
			     e3 = geo[i*width+(j+1)];

			Vec3 p1(e1.x, e1.y, e1.z),
			     p2(e2.x, e2.y, e2.z),
			     p3(e3.x, e3.y, e3.z);

			Vec3 v1 = p2-p1,
			     v2 = p3-p1;

			//            v1 = (p2[0]-p1[0], p2[1]-p1[1], p2[2]-p1[2]);
			//            v2 = (p3[0]-p1[0], p3[1]-p1[1], p3[2]-p1[2]);

			//Take the cross product
			Vec3 v3 = normalize(cross(v1, v2));
			size_t ind = i*(width-1)*3+j*3;
			//            nmap[ind] = (int)max(min(v3.x*256.+128., 255.),0.);
			//            nmap[ind+1] = (int)max(min(v3.y*256.+128., 255.),0.);
			//            nmap[ind+2] = (int)max(min(v3.z*256.+128., 255.),0.);
			nmap[ind] = (int)max(min(v3.x*128.+128., 255.),0.);
			nmap[ind+1] = (int)max(min(v3.y*128.+128., 255.),0.);
			nmap[ind+2] = (int)max(min(v3.z*128.+128., 255.),0.);
			//            Vec3 v3 = (v1[1]*v2[2]-v1[2]*v2[1], v1[2]*v2[0]-v1[0]*v2[2], v1[0]*v2[1]-v1[1]*v2[0]);
			//            lenv3 = sqrt(v3[0]**2 + v3[1]**2 + v3[2]**2);
			//            v3 = [v3[0]/lenv3, v3[1]/lenv3, v3[2]/lenv3];
			//            normalmap[j*3] = max(min(v3[0]*256+128, 255),0);
			//            normalmap[j*3+1] = max(min(v3[1]*256+128, 255),0);
			//            normalmap[j*3+2] = max(min(v3[2]*256+128, 255),0);
			//            normalmap.tofile(outputstream);
		}
	}
	printf("Done generating normal map.\n"); fflush(stdout);

    glCheckError("Failed to start creating normal texture");
	if (ntex)
		glDeleteTextures(1, &ntex);

	// Bind normalmap to a texture
	glGenTextures(1, &ntex);
	glBindTexture(GL_TEXTURE_2D, ntex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, nmap);

    glCheckError("Failed create normal texture");
    
    
    
    // Generate indices
    
    std::cout << "Width: " << width <<" Height: " << height << std::endl;
    
	std::vector<GLuint> indices;
	for(uint y = 0; y < height-1; ++y) {
		for(uint x = 0; x < width-1; ++x) {
			indices.push_back(y*width+x);
			if(x == 0 && y > 0)
				indices.push_back(indices.back());
			indices.push_back((y+1)*width+x);
		}
		if(y < height-2)
			indices.push_back(indices.back());
	}
	icount = (GLsizei)indices.size();
    std::cout << "icount: " << icount << std::endl;

	if (gvbo)
	{
		CUDA_UnregisterVBO(gvbo);
		glDeleteBuffers(1, &gvbo);
	}
	if (ivbo)
		glDeleteBuffers(1, &ivbo);
	
	
	// Bind the terrain coordinates to a VBO
	glGenBuffers(1, &gvbo);
	glBindBuffer(GL_ARRAY_BUFFER, gvbo);
	glBufferData(GL_ARRAY_BUFFER, 4 * size * sizeof(GLfloat), geo, GL_DYNAMIC_DRAW);	

	// Bind the indices to a VBO
	glGenBuffers(1, &ivbo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ivbo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, icount*sizeof(GLuint), &indices.front(), GL_DYNAMIC_DRAW);
	

    glCheckError("Error creating VAO");

    CUDA_RegisterVBO(gvbo);

     glCheckError("Error CUDA-Opengl terrain interop");
     
     //Create VAO
    glGenVertexArrays(1, &vba);
	glBindVertexArray(vba);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ivbo);
	glBindBuffer(GL_ARRAY_BUFFER, gvbo);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(GL_FLOAT), 0);
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);
     
	// Cleanup
	delete [] hmap;
	delete [] nmap;


	if (!snowtex)
	{
		// Load ground and snow textures from file using the SOIL library
		snowtex = SOIL_load_OGL_texture(
				"Snow.jpg",
				SOIL_LOAD_AUTO,
				SOIL_CREATE_NEW_ID,
				SOIL_FLAG_MIPMAPS | SOIL_FLAG_TEXTURE_REPEATS
				);
		
	}
	
	

	if (!groundtex)
	{
		groundtex = SOIL_load_OGL_texture(
				"grass.tga",
				SOIL_LOAD_AUTO,
				SOIL_CREATE_NEW_ID,
				SOIL_FLAG_MIPMAPS | SOIL_FLAG_TEXTURE_REPEATS
				);
	}

	// Bind the ground and snow textures
	glBindTexture(GL_TEXTURE_2D, snowtex);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 8.0f);
	
	glCheckError("Failed to bind snowtex");
	
	glBindTexture(GL_TEXTURE_2D, groundtex);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 8.0f);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	glCheckError("Failed to bind groundtex");

/*
	if (!hasShader)
	{
		shader.load("patch.vert", "patch.frag");
		hasShader = true;
	}
	*/
	drawShader_ = new ShaderProgram();
    drawShader_->loadShaderFromSource(GL_VERTEX_SHADER, "shaders/patch.glsl");
    drawShader_->loadShaderFromSource(GL_FRAGMENT_SHADER, "shaders/patch.glsl");
    drawShader_->link();

	glCheckError("Error loading shader");
	
	
}


void Terrain::Render() {
    
    //glDisable(GL_CULL_FACE);
   
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ntex);

	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, snowtex);

	glActiveTexture(GL_TEXTURE2);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, groundtex);

	//shader.bind();

	//shader.uniform("dim", SCENE_X);
	//shader.uniform("normalmap", 0);
	//shader.uniform("snowmap", 1);
	//shader.uniform("groundmap", 2);
	auto vsml = VSML::getInstance();
	
	drawShader_->bind();
	drawShader_->setUniformValue("dim", SCENE_X);
	drawShader_->setUniformValue("normalmap", 0);
	drawShader_->setUniformValue("snowmap", 1);
	drawShader_->setUniformValue("groundmap", 2);
	drawShader_->setUniformValue("terrain", 3);
	drawShader_->setUniformValue("projMatrix2", vsml->get(VSML::PROJECTION, false));
	drawShader_->setUniformValue("modelviewMatrix2", vsml->get(VSML::MODELVIEW, false));
    
    //glBindVertexArray(vba);
    //std::cout<<"Index size: " << icount << std::endl;
	//glDrawElements(GL_TRIANGLE_STRIP, icount, GL_UNSIGNED_INT, 0);
    
	//glBindVertexArray(0);

	//shader.unbind();
	
	quad_->draw(drawShader_);

	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE1);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE2);
	glDisable(GL_TEXTURE_2D);
}

void Terrain::UpdateNormalMap()
{
}



TessellationTerrain::TessellationTerrain(TerrainParams &params, Simulator *engine) {
    params_ = params;
    engine_ = engine;
    drawShader_ = new ShaderProgram();
    drawShader_->loadShaderFromSource(GL_VERTEX_SHADER, "shaders/recttess.glsl");
    drawShader_->loadShaderFromSource(GL_FRAGMENT_SHADER, "shaders/recttess.glsl");
    drawShader_->loadShaderFromSource(GL_GEOMETRY_SHADER, "shaders/recttess.glsl");
    drawShader_->loadShaderFromSource(GL_TESS_CONTROL_SHADER, "shaders/recttess.glsl");
    drawShader_->loadShaderFromSource(GL_TESS_EVALUATION_SHADER, "shaders/recttess.glsl");
    drawShader_->link();
    
    terrain_ = new Rect(glm::vec3(7, 0, 7), glm::vec3(0, 0, 0), glm::vec3(150, 1, 150));
    
    //quad used to render perlin noise to
    quad_ =  new Quad(glm::vec3(1, 1, 0), glm::vec3(params_.resolution * 0.5, params_.resolution * 0.5, 0), glm::vec3(params_.resolution, params_.resolution, 1));
    
    lod_ = 15.f;
    
    
    drawShader_->setFragDataLocation("out_Color0", 0);
    this->generateTerrain();
}

TessellationTerrain::~TessellationTerrain() {
    delete terrain_; 
    delete quad_;
    delete drawShader_;
    glDeleteTextures(1, &normalmap_);
    glDeleteTextures(1, &heightmap_);
    delete[] framebuffers_;
    delete[] heightmapData_;
}

void TessellationTerrain::generateTerrain() {
  
    // there be dragons ahead
    auto vsml = VSML::getInstance();
    int maxAttachments = 8;//GLFramebufferObject::queryMaxAttachments(); /* your platform better support 8 attachmenets or too bad */
    instances_ = params_.grid.x * params_.grid.y;
    int noBuffers = (int)ceilf(instances_ / (float)maxAttachments);
    framebuffers_ = new GLFramebufferObject*[noBuffers];
    
    //create the framebuffer
    
    GLFramebufferObjectParams params;
    params.width = params_.resolution;
    params.height = params_.resolution;
    
    params.hasDepth = false;
    params.nSamples = 0;
    params.type = GL_TEXTURE_3D;
    params.nColorAttachments = 0;
    params.format = GL_R16F;
    
    //create the 3d textire
    glGenTextures(1, &heightmap_);
    glBindTexture(GL_TEXTURE_3D, heightmap_);
    glTexParameterf(params.type, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(params.type, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
    glTexImage3D(GL_TEXTURE_3D, 0, GL_R16F, params_.resolution, params_.resolution, 
		 instances_, 0, GL_LUMINANCE, GL_FLOAT, 0);
    glBindTexture(GL_TEXTURE_3D, 0);

   
    
    
    // create lookup textures
    
    float permutation[256] = { 151,160,137,91,90,15,
    	131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,
    	190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,
    	88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,
    	77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,
    	102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,
    	135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,
    	5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,
    	223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,
    	129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,
    	251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,
    	49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
    	138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180
    };
    
    // gradients for 3d noise
    float g[] = {1,1,0,-1,1,0,1,-1,0,-1,-1,0,1,0,1,-1,0,1,1,0,-1,-1,0,-1,0,1,1,
		 0,-1,1,0,1,-1,	0,-1,-1,1,1,0,0,-1,1,-1,1,0,0,-1,-1};
    
    for(int i=0;i<256;i++) permutation[i] /= 255.0;

    GLuint textures[] = {0, 0};
    glGenTextures(2, &textures[0]);
    glBindTexture(GL_TEXTURE_1D, textures[0]);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexImage1D(GL_TEXTURE_1D, 0, GL_R16F, sizeof(permutation) / sizeof(float), 0, 
		 GL_LUMINANCE, GL_FLOAT, &permutation[0]);
    

    glBindTexture(GL_TEXTURE_1D, textures[1]);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB16F, sizeof(g) / sizeof(float) / 3, 0, 
		 GL_RGB, GL_FLOAT, &g[0]);
    glBindTexture(GL_TEXTURE_1D, 0);
    
    // create shaders
        
    perlinShader_ = new ShaderProgram();
    perlinShader_->loadShaderFromSource(GL_VERTEX_SHADER, "shaders/perlin.glsl");
    perlinShader_->loadShaderFromSource(GL_FRAGMENT_SHADER, "shaders/perlin.glsl");
    perlinShader_->link();
   
    
    // create quad
    Primitive *quad = new Quad(glm::vec3(1, 1, 0),glm::vec3(params.width * 0.5, params.height * 0.5, 0), glm::vec3(params.width, params.height, 1));
    
    // draw
    glm::vec2 random = params_.random;
    glm::vec2 *offsets = new glm::vec2[instances_ + (instances_ % 8)];
    for(int x=0, i=0; x<params_.grid.x; x++) {
	for(int y=0; y<params_.grid.y; y++, i++) {
	    offsets[i] = glm::vec2(x-1.0/params_.resolution*x,
				y-1.0/params_.resolution*y) ;
	    offsets[i].x += random.x;
	    offsets[i].y += random.y;
	}
    }
    
    int width = engine_->width();
    int height = engine_->height();
    
    glViewport(0, 0, params.width, params.height);
    engine_->setOrthoProjection(params_.resolution, params_.resolution);
    GLenum outputTex[8] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1,
			   GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3,
			   GL_COLOR_ATTACHMENT4, GL_COLOR_ATTACHMENT5,
			   GL_COLOR_ATTACHMENT6, GL_COLOR_ATTACHMENT7}; 
    // manually attach 3d layers to the framebuffers
    for(int i=0, k=0;i<noBuffers;i++) {
	    framebuffers_[i] = new GLFramebufferObject(params);
	    framebuffers_[i]->bind();
	    glBindTexture(GL_TEXTURE_3D, heightmap_);
	    
	    for(int j=0; j<maxAttachments; j++, k++) {
	        if(k >= instances_) 
	            break;
	        glFramebufferTextureLayer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+j, heightmap_, 0, k);
	    }
	    
	    glBindTexture(GL_TEXTURE_3D, 0);
	    framebuffers_[i]->release();
    }
   
    for(int i=0; i<noBuffers; i++) {
	    framebuffers_[i]->bind();
	    perlinShader_->bind();    
	    glActiveTexture(GL_TEXTURE0);
	    glBindTexture(GL_TEXTURE_1D, textures[0]);
	    perlinShader_->setUniformValue("permutation", 0);
	    glActiveTexture(GL_TEXTURE1);
	    glBindTexture(GL_TEXTURE_1D, textures[1]);
	    perlinShader_->setUniformValue("gradient", 1);
	    glActiveTexture(GL_TEXTURE0);
	    perlinShader_->setUniformValue("noiseScale", params_.noiseScale);
	    perlinShader_->setUniformValue("octaves", params_.octaves);
	    perlinShader_->setUniformValue("lacunarity", params_.lacunarity);
	    perlinShader_->setUniformValue("gain", params_.gain);
	    perlinShader_->setUniformValue("offset", params_.offset);
	    perlinShader_->setUniformValue("scale", params_.scale);
	    perlinShader_->setUniformValue("offsets", &offsets[8*i], 8);
	    perlinShader_->setUniformValue("projMatrix2", vsml->get(VSML::PROJECTION, false));
            perlinShader_->setUniformValue("modelviewMatrix2", vsml->get(VSML::MODELVIEW, false));
	    perlinShader_->setFragDataLocation("out_Color0", 0);
	    perlinShader_->setFragDataLocation("out_Color1", 1);
	    perlinShader_->setFragDataLocation("out_Color2", 2);
	    perlinShader_->setFragDataLocation("out_Color3", 3);
	    perlinShader_->setFragDataLocation("out_Color4", 4);
	    perlinShader_->setFragDataLocation("out_Color5", 5);
	    perlinShader_->setFragDataLocation("out_Color6", 6);
	    perlinShader_->setFragDataLocation("out_Color7", 7);
	    glDrawBuffers(8, outputTex); 
	    quad->draw(perlinShader_);
	    glDrawBuffers(1, outputTex); 
	    perlinShader_->release();
	    framebuffers_[i]->release();
    }
    
    heightmapData_ = new float[(int)(params_.resolution * params_.resolution * params_.grid.x * params_.grid.y)];
    
    glBindTexture(GL_TEXTURE_3D, heightmap_);
    glGetTexImage(GL_TEXTURE_3D, 0, GL_R16F, GL_FLOAT, &heightmapData_[0]);
    
    
    //create normal map
    glGenTextures(1, &normalmap_);
    glBindTexture(GL_TEXTURE_3D, normalmap_);
    glTexParameterf(params.type, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(params.type, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
    glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB16F, params_.resolution, params_.resolution, instances_, 0, GL_LUMINANCE, GL_FLOAT, 0);
    glBindTexture(GL_TEXTURE_3D, 0);


    for(int i=0, k=0;i<noBuffers;i++) {
	    framebuffers_[i]->bind();
	    glBindTexture(GL_TEXTURE_3D, normalmap_);
	    for(int j=0; j<maxAttachments; j++, k++) {
	        if(k >= instances_) 
	            break;
	        glFramebufferTextureLayer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+j, normalmap_, 0, k);
	    }
	    glBindTexture(GL_TEXTURE_3D, 0);
	    framebuffers_[i]->release();
    }
    
    lightingShader_ = new ShaderProgram();
    lightingShader_->loadShaderFromSource(GL_VERTEX_SHADER, "shaders/normals.glsl");
    lightingShader_->loadShaderFromSource(GL_FRAGMENT_SHADER, "shaders/normals.glsl");
    lightingShader_->link();
    
    float *layers = new float[instances_ + (instances_ % 8)];
    for(int x=0, i=0; x<params_.grid.x; x++) {
	    for(int y=0; y<params_.grid.y; y++, i++) {
	        layers[i] = (i+0.5) / (float)instances_;
	    }
    }

    for(int i=0; i<noBuffers; i++) { 
	    framebuffers_[i]->bind();
	
	    lightingShader_->bind();
	    glActiveTexture(GL_TEXTURE0);
	    glBindTexture(GL_TEXTURE_3D, heightmap_);
	    lightingShader_->setUniformValue("tex", 0);
	    lightingShader_->setUniformValue("layers", &layers[8*i], 8);
	    lightingShader_->setFragDataLocation("out_Color0", 0);
	    lightingShader_->setFragDataLocation("out_Color1", 1);
	    lightingShader_->setFragDataLocation("out_Color2", 2);
	    lightingShader_->setFragDataLocation("out_Color3", 3);
	    lightingShader_->setFragDataLocation("out_Color4", 4);
	    lightingShader_->setFragDataLocation("out_Color5", 5);
	    lightingShader_->setFragDataLocation("out_Color6", 6);
	    lightingShader_->setFragDataLocation("out_Color7", 7);
	    lightingShader_->setUniformValue("gird", params_.grid);
	    lightingShader_->setUniformValue("projMatrix2", vsml->get(VSML::PROJECTION, false));
	    lightingShader_->setUniformValue("modelviewMatrix2", vsml->get(VSML::MODELVIEW, false));
	    glDrawBuffers(8, outputTex); 
	    quad->draw(lightingShader_);
	
	    lightingShader_->release();
	    glBindTexture(GL_TEXTURE_3D, 0);
	    framebuffers_[i]->release();
	    delete framebuffers_[i];
    }
    
    glViewport(0, 0, width, height); // restore the viewport
    glBindTexture(GL_TEXTURE_1D, 0);
    glDeleteTextures(2, &textures[0]);
    delete quad;
    delete[] offsets;
    delete lightingShader_;

    
    GLTextureLoader::instance()->loadImage("textures/hello.bmp", "grass", GL_R11F_G11F_B10F, IL_RGB, IL_FLOAT);
    GLTextureLoader::instance()->loadImage("textures/sand.bmp", "sand", GL_R11F_G11F_B10F, IL_RGB, IL_FLOAT);
}

void TessellationTerrain::draw() {    
   
    auto vsml = VSML::getInstance();
    drawShader_->bind();
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_3D, heightmap_);
    drawShader_->setUniformValue("tex", 0);

    glActiveTexture(GL_TEXTURE2);
    glBindTexture(GL_TEXTURE_3D, normalmap_);
    drawShader_->setUniformValue("normalTex", 2);
    
    glActiveTexture(GL_TEXTURE3);
    glBindTexture(GL_TEXTURE_2D, (*GLTextureLoader::instance()->textures())["grass"].glTexId);
    drawShader_->setUniformValue("testTex", 3);
    
    glActiveTexture(GL_TEXTURE4);
    glBindTexture(GL_TEXTURE_2D, (*GLTextureLoader::instance()->textures())["sand"].glTexId);
    drawShader_->setUniformValue("sandTex", 4);
    
      
    drawShader_->setUniformValue("projMatrix2", vsml->get(VSML::PROJECTION, false));
    drawShader_->setUniformValue("modelviewMatrix2", vsml->get(VSML::MODELVIEW, false));
    
    drawShader_->setUniformValue("LOD", lod_);
    drawShader_->setUniformValue("lightPos", engine_->light());

    drawShader_->setUniformValue("cameraPos", engine_->getCamera()->eye);
    drawShader_->setUniformValue("grid", glm::vec2(params_.grid.x, params_.grid.y));
    drawShader_->setUniformValue("D", terrain_->scale().x);

    terrain_->draw(drawShader_, instances_);

}
