#ifndef TERRAIN_H
#define TERRAIN_H

#include <string>
#include <vector>
#include "vmath.h"
#include "opengl.h"
#include "defines.h"
#include "shader.h"
#include "Road.h"
#include "glm/glm.hpp"
#include <unordered_map>


class ShaderProgram;
class Primitive;
class GLFramebufferObject;
class VSML;
class Simulator;

class Terrain {
public:
    Terrain() : ivbo(0), gvbo(0), ntex(0), snowtex(0), groundtex(0), hasShader(false), geo(0) {}
	~Terrain();

	void Init(const std::string &hname, const std::string &nname, const uint width, const uint height, Road& road);

	void Render();
    void UpdateNormalMap();

	GLsizei icount;
	GLuint ivbo, gvbo, vba;
	GLuint ntex, ttex, snowtex, groundtex;
	Shader shader;
    bool hasShader;
    ShaderProgram *drawShader_;

private:
    Vec4* geo;
};


using namespace std;



struct TerrainParams {
    
    // geometry settings
    glm::vec3 tess;
    glm::vec2 scale;
    glm::vec2 grid;
    glm::vec2 random;
    
    // noise settings
    int resolution, octaves;
    float gain, lacunarity, offset, noiseScale;
};

class TessellationTerrain
{
public:
    TessellationTerrain(TerrainParams &params, Simulator *sim);
    ~TessellationTerrain();
    void draw ();
    float lod() { return lod_; }
    void setLod(float lod) {  lod_ = lod; }
    TerrainParams params() { return params_; }
protected:
    
    void generateTerrain();
    int instances_;
    GLuint heightmap_, normalmap_;
    TerrainParams params_;
    ShaderProgram *drawShader_, *perlinShader_, *lightingShader_;
    GLFramebufferObject **framebuffers_;
    Primitive *terrain_, *quad_;
    Simulator *engine_;
    GLuint vboPosID_, vaoID_;
    float *heightmapData_;
    float lod_;
};
#endif

