#include "Wind.h"
#include "WindSystem.cuh"

#include "CudaHelpers.h"

#include <cutil_inline.h>
#include <cutil_gl_inline.h>
#include <iostream>

Wind::Wind()
{
	has_allocated = false;
}

Wind::~Wind()
{
	if(has_allocated) {
                printf("Running wind sim desctructor.\n");
        	CUDA_WindField_Free();
        	glDeleteBuffers(1, &point_vbo);
                printf("Wind sim destructor done.\n");
	}
}


void Wind::Init(int xdim, int ydim, int zdim, unsigned int mapdim)
{
	has_allocated = true;
        std::cout << "Wind::Init()\n";

        dim3 dim;

        dim.x = xdim + 2;
        dim.y = ydim + 2;
        dim.z = zdim + 2;

        int size = dim.x * dim.y * dim.z;
        int size_bytes = size * sizeof(float4);

        point_count = size;

        CUDA_WindField_Init(xdim, ydim, zdim, mapdim);

        glGenBuffers(1, &point_vbo);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glBufferData(GL_ARRAY_BUFFER, size_bytes*2, 0, GL_DYNAMIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        CUDA_RegisterVBO(point_vbo);

	pressure_shader.load("pressure.vert", "pressure.frag");
        
	velocity_shader.load("velocity.vert", "velocity.frag");

        std::cout << "Wind::Init() done\n";
}

void Wind::SimulateWindField(float deltaTime, bool step, int frame, bool isRendering)
{
        CUDA_WindField_Simulate(deltaTime, step, frame, point_count, isRendering);
}

void Wind::RenderObstacles()
{
        CUDA_WindField_RenderObstacles(point_vbo, point_count);

        glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glVertexPointer(4, GL_FLOAT, 0, 0);
        glDrawArrays(GL_POINTS, 0, point_count);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void Wind::RenderPressure()
{
        CUDA_WindField_RenderPressure(point_vbo, point_count);

        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        pressure_shader.bind();
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glVertexPointer(4, GL_FLOAT, 0, 0);
        glDrawArrays(GL_POINTS, 0, point_count);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        pressure_shader.unbind();
        glDisable(GL_BLEND);
        glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);

}

void Wind::RenderDebug(float limit)
{
        CUDA_WindField_RenderDebug(limit, point_vbo, point_count);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        velocity_shader.bind();
        glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glVertexPointer(4, GL_FLOAT, 0, 0);
        glDrawArrays(GL_LINES, 0, point_count*2);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        velocity_shader.unbind();
        glDisable(GL_BLEND);
}

