#ifndef WINDLBM_CPU_H
#define WINDLBM_CPU_H

#include <ctime>

#include <sstream>

// C library to perform input and output operations
#include <stdio.h>

// C standard general utilities library 
#include <stdlib.h>

// The OpenGL Extension Wrangler Library 
#include <GL/glew.h>

// The OpenGL Utility Toolkit
#include <GL/glut.h>

//#include "LBM_config.h"
#include "LBM_types.h"

#include "shader.h"
#include "defines.h"

#include <iostream>
#include <string>

// Timing on Linux
//#include <sys/time.h>

using namespace std;

class WindLBM
{
public:
	WindLBM();
	~WindLBM(); 

/**
 * Public functions
 */
        void Init(int sizeX, int sizeY, int sizeZ, unsigned int mapDim, float_type lbmOmega, float_type lbmViscosity, float4 velFactor, float_type initVel, float_type initRho, float4 boundaryVelFactor);       
	// Get physical dimension in x 
	int getSizeX();

	// Get physical dimension in y
	int getSizeY();

	// Get physical dimension in z
	int getSizeZ();
	
	// Get graphics rendering on/off
	bool graphics();

	// Set graphics rendering on/off
	void doGraphics();

        // Get current iteration number	
	int getIteration();
	
        // Prints the initial state to console
	void printInitalState();

        // Snow simulator integration functions
        void SimulateWindField(int collideKernelType, float deltaTime, bool step, int frame, bool isRendering, float& mlups);

        // Debug rendering
        void RenderObstacles();
        void RenderPressure();
	void RenderDebug(float limit);
	
private:

/**
 *  Private functions
 */

	// Allocate memory for the CPU
	void allocateMemoryCPU();

	// Get index at location x, y, and z
	int getIndex(int x, int y, int z);

	void initData();

/**		
 * Variables
 */

	// Boundary velocities
        int num_lbm_speeds;
	float4 lbm_speeds[4];
	int lbm_speed_idx;

        float4 velocity_factor;
        float4 boundary_velocity_factor;

	float total_sim_time;
	
	int boundary_size;

        // Used for field shaders
        GLuint point_vbo, height_vbo;
        GLsizei point_count;
	
        Shader pressure_shader, velocity_shader;        

	bool has_allocated;

	// Force to add to each lattice location
	float_type force;

	// Physical dimension in x
	int size_x;

	// Physical dimension in y
	int size_y;

	// Physical dimension in z
	int size_z;

	// Relaxation parameter
	float_type omega;
	
        float_type initial_rho;

	// Host memory for solids 
	uchar *solid_h; 

	// Host memory for the particle distribution functions
	float_type *f0_h, *f1_h, *f2_h, *f3_h, *f4_h, *f5_h, *f6_h, *f7_h, *f8_h, *f9_h, 
		  *f10_h, *f11_h, *f12_h, *f13_h, *f14_h, *f15_h, *f16_h, *f17_h, *f18_h;
	
	// Viscosity is the quantity that describes a fluid's resistance to flow
	double viscosity;

	// Extra space around the fluid lattice
	int extra;

	// Render graphics
	bool render_graphics; 
	
	// The solid boundaries configuration
	int boundaries[3];

        int iter_count;
};

#endif
