#ifndef WINDSYSTEMKERNELS_H
#define WINDSYSTEMKERNELS_H


//texture<float4, 3> wind_vel_tex;
cudaArray *wind_vel_array;

cudaPitchedPtr wind_vel;
cudaPitchedPtr pressure;
cudaPitchedPtr pressureB;
cudaPitchedPtr solution;
cudaPitchedPtr obstacle;
//cudaPitchedPtr curlmag;

cudaExtent wind_extent;
cudaExtent pressure_extent;
cudaExtent pressureB_extent;
cudaExtent obstacle_extent;

#define NUM_SPEEDS 4
static float4 speeds[] = {
    make_float4(1.0f, 0.0f, 0.0f, 1.0f),
    make_float4(3.0f, 0.0f, 1.0f, 1.0f),
    make_float4(1.0f, 0.0f, -2.0f, 1.0f),
    make_float4(2.0f, 0.0f, 0.0f, 1.0f),
};

#define BLOCK_X 16
#define BLOCK_Y 3
#define BLOCK_Z 16
#define BLOCK_W (BLOCK_X+2)
#define BLOCK_H (BLOCK_Y+2)
#define BLOCK_D (BLOCK_Z+2)
#define BLOCK_SIZE (BLOCK_W*BLOCK_D)
//#define BLOCK_SIZE (BLOCK_W*BLOCK_H*BLOCK_D)

//#define STENCIL(i,j) block[BLOCK_W*((int)threadIdx.y+1+(j))+(int)threadIdx.x+1+(i)]
#define STENCIL3(i,j,k) block[BLOCK_W*BLOCK_D*((int)threadIdx.y+1+(j)) + BLOCK_W*((int)threadIdx.z+1+(k)) + (int)threadIdx.x+1+(i)]

#define XBORDER(Y) *((float *)((char *)q.ptr + q.pitch*q.ysize*(Y) + q.pitch*blockIdx.x) \
                                        + blockIdx.y*BLOCK_X + threadIdx.x)

#define LOAD(Y) \
        block[0] = *(float *)((char *)p.ptr + p.ysize*p.pitch*(Y)); \
        if(blockIdx.y == 0) { \
                if(threadIdx.y == 0) block[-BLOCK_W] = block[0]; \
                if(threadIdx.y == BLOCK_Z-1) block[BLOCK_W] = block[0]; \
        } else { \
                if(threadIdx.y == 0) block[-BLOCK_W] = *(float *)((char *)p.ptr + p.ysize*p.pitch*(Y) - p.pitch); \
                if(threadIdx.y == BLOCK_Z-1) block[BLOCK_W] = *(float *)((char *)p.ptr + p.ysize*p.pitch*(Y) + p.pitch); \
        } \
        if(threadIdx.x == 0) block[-1] = block[0]; \
        if(threadIdx.x == BLOCK_X-1) block[1] = block[0];

#define LOAD3(Y) \
        STENCIL3(0,0,0) = REF(float, p, x, (Y), z); \
        if(threadIdx.z == 0) STENCIL3(0,0,-1) = STENCIL3(0,0,0); \
        if(threadIdx.z == BLOCK_Z-1) STENCIL3(0,0,1) = STENCIL3(0,0,0); \
        if(threadIdx.y == 0) STENCIL3(0,-1,0) = STENCIL3(0,0,0); \
        if(threadIdx.y == BLOCK_Y-1) STENCIL3(0,1,0) = STENCIL3(0,0,0); \
        if(threadIdx.x == 0) STENCIL3(-1,0,0) = STENCIL3(0,0,0); \
        if(threadIdx.x == BLOCK_X-1) STENCIL3(1,0,0) = STENCIL3(0,0,0);
/*
        if(threadIdx.y < 2) { \
                int off = threadIdx.y * BLOCK_Z - (1 - threadIdx.y); \
                STENCIL(0,off) = REF(float, p, x, (Y), blockIdx.y*BLOCK_Z+1+off); \
        } \
        if(threadIdx.y < 1) { \
                int off = (threadIdx.x >> 3) * (BLOCK_X+1); \
                int yy = threadIdx.x & 7; \
                block[(BLOCK_X+2)*(yy+1)+off] = XBORDER((Y)); \
        }*/

#endif
