#ifndef WINDKERNEL_WINDLBM_H
#define WINDKERNEL_WINDLBM_H

#include <cutil.h>
#include <cuda_runtime_api.h>
#include <stdio.h>

//#define REF(T, p, x, y, z) (*((T *)((char *)(p).ptr + __mul24(__mul24((p).pitch, (p).ysize), (z)) + __mul24((p).pitch, (y))) + (x)))

	//#define REF_WIND(T,p,x,y,z) (*((T *)((char *)(p).ptr + ((p).pitch * (p).ysize * (y)) + (p).pitch * (z)) + (x)))
	//#define OBS(obstacles, x, y, z) ((*((int *)((char *)(obstacles).ptr + ( (obstacles).pitch * (obstacles).ysize * (z)) + (obstacles).pitch * (y)) + (x))) & 1 )
	//#define OBS_REF(T,p,x,y,z) (*((T *)((char *)(p).ptr + ( (p).pitch * (p).ysize * (y)) + (p).pitch * (z)) + (x)))
	//#define PDF_REF(T, p, x, y, z, pitch, ysize) (*((T *)((char *)p + (pitch * ysize * (z)) + pitch * (y)) + (x)))

//#ifdef FERMI_OPT
//	#define _MUL(A,B) A*B
//#else
//	#define _MUL(A,B) __mul24(A,B)
//#endif
//*
#define REF_WIND(T,p,x,y,z) (*((T *)((char *)(p).ptr + __mul24(__mul24((p).pitch, (p).ysize), (y)) + __mul24((p).pitch, (z))) + (x)))

// Retrieves a value from a snow simulation obstacle
// True value if the location is an obstacle
// -- REF(int, obstacles, x, y, z) & 1 --
#define OBS(obstacles, x, y, z) ((*((int *)((char *)(obstacles).ptr + __mul24(__mul24((obstacles).pitch, (obstacles).ysize), (z)) + __mul24((obstacles).pitch, (y))) + (x))) & 1 )
#define OBS_REF(T,p,x,y,z) (*((T *)((char *)(p).ptr + __mul24(__mul24((p).pitch, (p).ysize), (y)) + __mul24((p).pitch, (z))) + (x)))
#define PDF_REF(T, p, x, y, z, pitch, ysize) (*((T *)((char *)p + __mul24(__mul24(pitch, ysize), (z)) + __mul24(pitch, (y))) + (x)))
//*/
struct PDF
{
	void * ptr[19];
};

PDF pdf;

extern __shared__ char shared[];

__constant__ float t_3d_dev[19] = 
{ 
	1./3.,  // 0
	1./18., // 1
	1./18., // 2
	1./18., // 3
	1./36., // 4
	1./36., // 5
	1./36., // 6
	1./36., // 7
	1./36., // 8
	1./36., // 9      
	1./18., // 10
	1./18., // 11
	1./18., // 12         
	1./36., // 13
	1./36., // 14
	1./36., // 15   
	1./36., // 16
	1./36., // 17
	1./36.  // 18
};

__constant__ float c_3d_dev[19][3] = 
{
	{0, 0, 0},    // 0
	{-1, 0, 0},   // 1
	{0, -1, 0},   // 2
	{0, 0, -1},   // 3
	{-1, -1, 0},  // 4
	{-1, 1, 0},   // 5
	{-1, 0, -1},  // 6 
	{-1, 0, 1},   // 7 
	{0, -1, -1},  // 8
	{0, -1, 1},   // 9
	{1, 0, 0},    // 10
	{0, 1, 0},    // 11
	{0, 0, 1},    // 12
	{1, 1, 0},    // 13
	{1, -1, 0},   // 14
	{1, 0, 1},    // 15
	{1, 0, -1},   // 16
	{0, 1, 1},    // 17
	{0, 1, -1}    // 18
};

// Variables in constant memory 
__constant__ int lbm_ysize_dev[1];
__constant__ int lbm_pitch_dev[1];
__constant__ int lbm_boundary_size_dev[1];
__constant__ float4 lbm_vel_factor_dev[1];

int* numActiveFluidCells;

cudaPitchedPtr f_d[19];	

// Device memory for solids 
float_type *solid_d;

// Device memory for the particle distribution functions
float_type *f0_d, *f1_d, *f2_d, *f3_d, *f4_d, *f5_d, *f6_d, *f7_d, *f8_d, *f9_d, *f10_d, *f11_d, *f12_d, *f13_d, *f14_d, *f15_d, *f16_d, *f17_d, *f18_d;

// Physical dimension in x, y, and z
int dimX, dimY, dimZ;

/**
 * Make velocity lines (for debug rendering). 
 * Samples a texture of the velocity field.
 */
__global__ void make_vel_lines_lbm(cudaPitchedPtr vel, cudaPitchedPtr obs, Line *out, float limit, float3 idim) {
        int tid = __mul24(blockIdx.x, blockDim.x) + threadIdx.x;


		int x = threadIdx.x;
        int y = blockIdx.x / obs.ysize;
        int z = blockIdx.x - y * obs.ysize;

        volatile float4 v = REF_WIND(float4, vel, x, y, z);
        //int mask = REF_WIND(int, obs, x, y, z) & 127;
        
        float speed = sqrtf(v.x*v.x + v.y*v.y + v.z*v.z);
        float temp = speed;

	if(temp < 1.0f)
		temp = 1.0f;
	else
		temp = 1.0f / temp;

        //if(!mask) temp = 0.0f;

        Line l;

        l.p0.x = (float)x;
        l.p0.y = (float)y;
        l.p0.z = (float)z;
        l.p0.w = 1.0f;

        l.p1.x = l.p0.x + v.x*temp;
        l.p1.y = l.p0.y + v.y*temp;
        l.p1.z = l.p0.z + v.z*temp;
        l.p1.w = 1.0f;

        l.p0.x *= idim.x;
        l.p1.x *= idim.x;
        l.p0.y *= idim.y;
        l.p1.y *= idim.y;
        l.p0.z *= idim.z;
        l.p1.z *= idim.z;

        //volatile float4 vv = wind_vel_sample(l.p1);
        float4 vv = wind_vel_sample(l.p1);
        float speed2 = sqrtf(vv.x*vv.x + vv.y*vv.y + vv.z*vv.z);
        l.p0.w = l.p1.w = speed2 - speed;

        out[tid] = l;
}

/** 
 * Kernel - Get Rho 2
 */
__device__ void getRhoRegular(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type &rho)
{
	 int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	 #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
		
	 rho = PDF_VAL(4) + PDF_VAL(5) + PDF_VAL(6) + PDF_VAL(7) +  PDF_VAL(8) + PDF_VAL(9) + 
	       PDF_VAL(13) + PDF_VAL(14) + PDF_VAL(15) + PDF_VAL(16) + PDF_VAL(17) + PDF_VAL(18) +
	       PDF_VAL(1) + PDF_VAL(2) + PDF_VAL(3) + PDF_VAL(11) + PDF_VAL(12) + PDF_VAL(10) +  PDF_VAL(0);  	       
}

/**
 * Kernel - Get Velocity X 2
 */
__device__ void getVelocityXRegular(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type rho, float_type &ux)
{
	 int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	 #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
	ux = (
	     (PDF_VAL(13) - PDF_VAL(4)) 
	   + (PDF_VAL(14) - PDF_VAL(5)) 
	   + (PDF_VAL(15) - PDF_VAL(6)) 
	   + (PDF_VAL(16) - PDF_VAL(7))
	   + (PDF_VAL(10) - PDF_VAL(1)) 
	   ) / rho;	
}

/**
 * Kernel - Get Velocity Y 2
 */
__device__ void getVelocityYRegular(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type rho, float_type &uy)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
     uy = (
	     (PDF_VAL(5) - PDF_VAL(14)) 
	   + (PDF_VAL(11) - PDF_VAL(2)) 
	   + (PDF_VAL(13) - PDF_VAL(4)) 
	   + (PDF_VAL(17) - PDF_VAL(8)) 
	   + (PDF_VAL(18) - PDF_VAL(9))) / rho;       
}

/**
 * Kernel - Get Velocity Z 2
 */
__device__ void getVelocityZRegular(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type rho, float_type &uz)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

     uz = (
	     (PDF_VAL(7) - PDF_VAL(16)) 
	   + (PDF_VAL(9) - PDF_VAL(18)) 
	   + (PDF_VAL(15) - PDF_VAL(6)) 
	   + (PDF_VAL(17) - PDF_VAL(8))
	   + (PDF_VAL(12) - PDF_VAL(3)) 
	   ) / rho;
}

/**
 * Kernel - Get Rho 2
 */
__device__ void getRho2(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type 
&rho)
{
    int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
       #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
       rho = PDF_VAL(0) + PDF_VAL(1) + PDF_VAL(2) + PDF_VAL(3) + PDF_VAL(4) + PDF_VAL(5) +
          PDF_VAL(6) + PDF_VAL(7) + PDF_VAL(8) + PDF_VAL(9) + PDF_VAL(10) + PDF_VAL(11) +
          PDF_VAL(12) + PDF_VAL(13) + PDF_VAL(14) +  PDF_VAL(15) + PDF_VAL(16) + 
PDF_VAL(17) + PDF_VAL(18);             }

/**
 * Kernel - Get Velocity X 2
 */
__device__ void getVelocityX2(PDF pdf, int x, int y, int z, size_t pitch, int height, 
float_type rho, float_type &ux)
{
    int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
       #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
       ux = (PDF_VAL(10) + PDF_VAL(13) + PDF_VAL(14) + PDF_VAL(15) + PDF_VAL(16) - 
(PDF_VAL(1) + PDF_VAL(4) + PDF_VAL(5) + PDF_VAL(6) + PDF_VAL(7))) / rho;                  
    }

/**
 * Kernel - Get Velocity Y 2
 */
__device__ void getVelocityY2(PDF pdf, int x, int y, int z, size_t pitch, int height, 
float_type rho, float_type &uy)
{
   int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
      #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
      uy = (PDF_VAL(5) + PDF_VAL(11) +  PDF_VAL(13) + PDF_VAL(17) + PDF_VAL(18) - 
(PDF_VAL(2) + PDF_VAL(4) + PDF_VAL(8) + PDF_VAL(9) + PDF_VAL(14))) / rho;           }

/**
 * Kernel - Get Velocity Z 2
 */
__device__ void getVelocityZ2(PDF pdf, int x, int y, int z, size_t pitch, int height, 
float_type rho, float_type &uz)
{
   int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
      #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
      uz = (PDF_VAL(7) + PDF_VAL(9) + PDF_VAL(12) + PDF_VAL(15) + PDF_VAL(17) - 
(PDF_VAL(3) + PDF_VAL(6) + PDF_VAL(8) + PDF_VAL(16) + PDF_VAL(18))) / rho;
} 

/**
 * Kernel - Get Rho 5
 */
__device__ void getRhoPrecise(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type &delta_rho)
{
	 int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	 #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

	 delta_rho = PDF_VAL(4) + PDF_VAL(5) + PDF_VAL(6) + PDF_VAL(7) +  PDF_VAL(8) + PDF_VAL(9) + 
	             PDF_VAL(13) + PDF_VAL(14) + PDF_VAL(15) + PDF_VAL(16) + PDF_VAL(17) + PDF_VAL(18) +
	             PDF_VAL(1) + PDF_VAL(2) + PDF_VAL(3) + PDF_VAL(11) + PDF_VAL(12) + PDF_VAL(10) +  PDF_VAL(0);  	       
}

/**
 * Kernel - Get Velocity X 5
 */
__device__ void getVelocityXPrecise(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type rho, float_type &ux)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
	ux = (
	     (PDF_VAL(13) - PDF_VAL(4)) 
	   + (PDF_VAL(14) - PDF_VAL(5)) 
	   + (PDF_VAL(15) - PDF_VAL(6)) 
	   + (PDF_VAL(16) - PDF_VAL(7))
	   + (PDF_VAL(10) - PDF_VAL(1)) 
	   ) / rho;
}

/**
 * Kernel - Get Velocity Y 5
 */
__device__ void getVelocityYPrecise(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type rho, float_type &uy)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
	uy = (
	     (PDF_VAL(5) - PDF_VAL(14)) 
	   + (PDF_VAL(11) - PDF_VAL(2)) 
	   + (PDF_VAL(13) - PDF_VAL(4)) 
	   + (PDF_VAL(17) - PDF_VAL(8)) 
	   + (PDF_VAL(18) - PDF_VAL(9))) / rho;    
}

/**
 * Kernel - Get Velocity Z 5
 */
__device__ void getVelocityZPrecise(PDF pdf, int x, int y, int z, size_t pitch, int height, float_type rho, float_type &uz)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

	uz = (
	     (PDF_VAL(7) - PDF_VAL(16)) 
	   + (PDF_VAL(9) - PDF_VAL(18)) 
	   + (PDF_VAL(15) - PDF_VAL(6)) 
	   + (PDF_VAL(17) - PDF_VAL(8))
	   + (PDF_VAL(12) - PDF_VAL(3)) 
	   ) / rho;
}

/**
 * Kernel - Get Rho
 */
__device__ void getRho5Double(PDF pdf, int x, int y, int z, size_t pitch, int height, double &delta_rho)
{
        int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
        #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
	delta_rho = PDF_VAL(4) + PDF_VAL(5) + PDF_VAL(6) + PDF_VAL(7) +  PDF_VAL(8) + PDF_VAL(9) +  PDF_VAL(13) + PDF_VAL(14) + PDF_VAL(15) + PDF_VAL(16) + PDF_VAL(17) + PDF_VAL(18) +
	      PDF_VAL(1) + PDF_VAL(2) + PDF_VAL(3) + PDF_VAL(11) + PDF_VAL(12) + PDF_VAL(10) +  PDF_VAL(0);  	       
}

/**
 * Kernel - Get Velocity X 5
 */
__device__ void getVelocityX5Double(PDF pdf, int x, int y, int z, size_t pitch, int height, double rho, double &ux)
{
	 int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	 #define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
	ux = (
	     (PDF_VAL(13) - PDF_VAL(4)) 
	   + (PDF_VAL(14) - PDF_VAL(5)) 
	   + (PDF_VAL(15) - PDF_VAL(6)) 
	   + (PDF_VAL(16) - PDF_VAL(7))
	   + (PDF_VAL(10) - PDF_VAL(1)) 
	   ) / rho;
}

/**
 * Kernel - Get Velocity Y 5
 */
__device__ void getVelocityY5Double(PDF pdf, int x, int y, int z, size_t pitch, int height, double rho, double &uy)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
     uy = (
	     (PDF_VAL(5) - PDF_VAL(14)) 
	   + (PDF_VAL(11) - PDF_VAL(2)) 
	   + (PDF_VAL(13) - PDF_VAL(4)) 
	   + (PDF_VAL(17) - PDF_VAL(8)) 
	   + (PDF_VAL(18) - PDF_VAL(9))) / rho; 
}

/**
 * Kernel - Get Velocity Z 5
 */
__device__ void getVelocityZ5Double(PDF pdf, int x, int y, int z, size_t pitch, int height, double rho, double &uz)
{
	int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, height);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

     uz = (
	     (PDF_VAL(7) - PDF_VAL(16)) 
	   + (PDF_VAL(9) - PDF_VAL(18)) 
	   + (PDF_VAL(15) - PDF_VAL(6)) 
	   + (PDF_VAL(17) - PDF_VAL(8))
	   + (PDF_VAL(12) - PDF_VAL(3)) 
	   ) / rho;
}

/**
 * Kernel - Create Velocity Field - first variant, precise
 *           Follows same pattern as createLines_Kernel(2)
 */
__global__ void createVelocityField_Kernel(cudaPitchedPtr solid_d, cudaPitchedPtr vel_field, PDF pdf, const int dimX, const int dimY, const int dimZ)
{	
	int x = threadIdx.x;
        int grid_y = __mul24(blockIdx.y, gridDim.x);
        int z = (blockIdx.x + grid_y) / dimY;
        int y = (blockIdx.x + grid_y) - z * dimY;

        /*int x = blockIdx.x;
        int grid_y = __mul24(blockIdx.y, blockDim.x);
        int z = (threadIdx.x + grid_y) / blockDim.x;
        int y = (threadIdx.x + grid_y) - z * blockDim.x;*/

	float_type rho, ux, uy, uz;
	float4 u;

	if ((x < dimX) && (y < dimY) && (z < dimZ))
	{
		//if(OBS_REF(int, solid_d, x, y, z) == 0) {
			// Compute density from the particle distribution functions
			getRhoPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho);
			getVelocityXPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, ux);
			getVelocityYPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, uy);
			getVelocityZPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, uz);
			u.x = ux*14; u.y = uy*14; u.z = uz*14;
			//__syncthreads();
			//REF_WIND(float4, vel_field, x, y, z) = u;
		/*}
                else {
                        u.x = 0.0;
                        u.y = 0.0;
                        u.z = 0.0;
                }*/
                REF_WIND(float4, vel_field, x, y, z) = u;
	}
}

/**
 * Kernel - Create Velocity Field - second variant, precise
 *           Follows same pattern as createLines_Kernel(2)
 */
__global__ void createVelocityField_Kernel2(cudaPitchedPtr solid_d, cudaPitchedPtr vel_field, PDF pdf, const int dimX, const int dimY, const int dimZ)
{	
        int x = threadIdx.x;
        int y = blockIdx.x;
        int z = blockIdx.y; 

	float_type rho, ux, uy, uz;
	float4 u;

	if ((x < dimX) && (y < dimY) && (z < dimZ))
	{
		if(OBS_REF(int, solid_d, x, y, z) == 0) {
			// Compute density from the particle distribution functions
			getRhoPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho);
			getVelocityXPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, ux);
			getVelocityYPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, uy);
			getVelocityZPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, uz);
			u.x = ux*lbm_vel_factor_dev->x; u.y = uy * lbm_vel_factor_dev->y; u.z = uz * lbm_vel_factor_dev->z;
		}
                else {
                        u.x = 0.0;
                        u.y = 0.0;
                        u.z = 0.0;
                }
                REF_WIND(float4, vel_field, x, y, z) = u;
	}
}

/**
 * Generate pressure field
 * Presently does not work
 */
__global__ void createPressureField_Kernel(float4 *out, cudaPitchedPtr solid_d, PDF pdf, dim3 dim, float3 idim) 
{
        int x = threadIdx.x;
        int grid_y = __mul24(blockIdx.y, gridDim.x);
        int z = (blockIdx.x + grid_y) / dim.y;
        int y = (blockIdx.x + grid_y) - z * dim.y;

	int tid = x + y + z;

	float_type rho, ux, uy, uz, val;

	if ((x < dim.x) && (y < dim.y) && (z < dim.z))
        {
                if(OBS_REF(int, solid_d, x, y, z) == 0) {
                        getRhoPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho);
                        getVelocityXPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, ux);
                        getVelocityYPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, uy);
                        getVelocityZPrecise(pdf, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev, rho+19.f, uz);
			// http://en.wikipedia.org/wiki/Dynamic_pressure
			val = 0.5 * rho * (ux*10 * ux*10 + uy*10 * uy*10 + uz*10 * uz*10);
			out[tid] = make_float4((float)x*idim.x, (float)y*idim.y, (float)z*idim.z, val);

                }
        }
	
        out[tid] = make_float4((float)x*idim.x, (float)y*idim.y, (float)z*idim.z, val);
}

#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

// Macro to set the PDFs for a given lattice position to a given velocity (and rho)
// Useful for initializing and boundaries
// Uses the same equilibrium calculations as in the "precise" kernels
// Exists because of lacking device function functionality
#define SET_PDFS_TO_VEL_PRECISE(pdf, x_pos, y_pos, z_pos, delta_rho, ux, uy, uz) \
        volatile int offset = (int) &PDF_REF(float_type, 0, x_pos, y_pos, z_pos, *lbm_pitch_dev, *lbm_ysize_dev); \
        float_type rho = delta_rho + 19.f; \
        float_type uSqr = ux*ux + uy*uy + uz*uz; \
        float_type u_sqr = 1.5f * uSqr; \
        float_type a = 4.5f; \
        \
        float_type cu; \
        \
        PDF_VAL(0) = t_3d_dev[0]*delta_rho - t_3d_dev[0]*(rho)*u_sqr; \
        cu = c_3d_dev[1][0]*ux + c_3d_dev[1][1]*uy + c_3d_dev[1][2]*uz; \
        PDF_VAL(1) = t_3d_dev[1]*delta_rho + t_3d_dev[1]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[2][0]*ux + c_3d_dev[2][1]*uy + c_3d_dev[2][2]*uz; \
        PDF_VAL(2) = t_3d_dev[2]*delta_rho + t_3d_dev[2]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[3][0]*ux + c_3d_dev[3][1]*uy + c_3d_dev[3][2]*uz; \
        PDF_VAL(3) = t_3d_dev[3]*delta_rho + t_3d_dev[3]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[4][0]*ux + c_3d_dev[4][1]*uy + c_3d_dev[4][2]*uz; \
        PDF_VAL(4) = t_3d_dev[4]*delta_rho + t_3d_dev[4]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[5][0]*ux + c_3d_dev[5][1]*uy + c_3d_dev[5][2]*uz; \
        PDF_VAL(5) = t_3d_dev[5]*delta_rho + t_3d_dev[5]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[6][0]*ux + c_3d_dev[6][1]*uy + c_3d_dev[6][2]*uz; \
        PDF_VAL(6) = t_3d_dev[6]*delta_rho + t_3d_dev[6]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[7][0]*ux + c_3d_dev[7][1]*uy + c_3d_dev[7][2]*uz; \
        PDF_VAL(7) = t_3d_dev[7]*delta_rho + t_3d_dev[7]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[8][0]*ux + c_3d_dev[8][1]*uy + c_3d_dev[8][2]*uz; \
        PDF_VAL(8) = t_3d_dev[8]*delta_rho + t_3d_dev[8]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[9][0]*ux + c_3d_dev[9][1]*uy + c_3d_dev[9][2]*uz; \
        PDF_VAL(9) = t_3d_dev[9]*delta_rho + t_3d_dev[9]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[10][0]*ux + c_3d_dev[10][1]*uy + c_3d_dev[10][2]*uz; \
        PDF_VAL(10) = t_3d_dev[10]*delta_rho + t_3d_dev[10]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[11][0]*ux + c_3d_dev[11][1]*uy + c_3d_dev[11][2]*uz; \
        PDF_VAL(11) = t_3d_dev[11]*delta_rho + t_3d_dev[11]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[12][0]*ux + c_3d_dev[12][1]*uy + c_3d_dev[12][2]*uz; \
        PDF_VAL(12) = t_3d_dev[12]*delta_rho + t_3d_dev[12]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[13][0]*ux + c_3d_dev[13][1]*uy + c_3d_dev[13][2]*uz; \
        PDF_VAL(13) = t_3d_dev[13]*delta_rho + t_3d_dev[13]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[14][0]*ux + c_3d_dev[14][1]*uy + c_3d_dev[14][2]*uz; \
        PDF_VAL(14) = t_3d_dev[14]*delta_rho + t_3d_dev[14]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[15][0]*ux + c_3d_dev[15][1]*uy + c_3d_dev[15][2]*uz; \
        PDF_VAL(15) = t_3d_dev[15]*delta_rho + t_3d_dev[15]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[16][0]*ux + c_3d_dev[16][1]*uy + c_3d_dev[16][2]*uz; \
        PDF_VAL(16) = t_3d_dev[16]*delta_rho + t_3d_dev[16]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[17][0]*ux + c_3d_dev[17][1]*uy + c_3d_dev[17][2]*uz; \
        PDF_VAL(17) = t_3d_dev[17]*delta_rho + t_3d_dev[17]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr); \
        \
        cu = c_3d_dev[18][0]*ux + c_3d_dev[18][1]*uy + c_3d_dev[18][2]*uz; \
        PDF_VAL(18) = t_3d_dev[18]*delta_rho + t_3d_dev[18]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);


/**
 * Kernel - Apply outer boundaries
 */
__global__ void applyOuterBoundariesGPU_kernel(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ, float3 inVel, float inRho)
{
        int x = threadIdx.x;
        int y = blockIdx.x / solid_d.ysize;
        int z = blockIdx.x - __mul24(y, solid_d.ysize);

        int mask = OBS_REF(int, solid_d, x, y, z);
        #if PRINT_DEBUG_KERNEL
                if(x == 0 && y == 0 && z == 0)
                        printf("Outer boundaries kernel: Received invel %f, %f, %f\n", inVel.x, inVel.y, inVel.z);
        #endif

        if(x < *lbm_boundary_size_dev || x >= (dimX - *lbm_boundary_size_dev) || 
	   y < *lbm_boundary_size_dev || y >= (dimY - *lbm_boundary_size_dev) ||
	   z < *lbm_boundary_size_dev || z >= (dimZ - *lbm_boundary_size_dev)) {
	/*if(x < *lbm_boundary_size_dev || x >= (dimX - *lbm_boundary_size_dev) || 
	   z < *lbm_boundary_size_dev || z >= (dimZ - *lbm_boundary_size_dev)) {*/
                SET_PDFS_TO_VEL_PRECISE(pdf, x, y, z, inRho, inVel.x, inVel.y, inVel.z)
	}
        // Outflow	
        /*else if(x >= (dimX - *lbm_boundary_size_dev)) {
         	#define PDF_VAL2(n) (*((float_type*) ((char*)pdf.ptr[n] + offset2)))               
                volatile int offset = (int) &PDF_REF(float_type, 0, x, y, z, f_pitch, ysize);
                volatile int offset2 = (int) &PDF_REF(float_type, 0, x, y, z, f_pitch, ysize) - 1;
 		PDF_VAL(14) = PDF_VAL2(14);
		PDF_VAL(2) = PDF_VAL2(2);
		PDF_VAL(4) = PDF_VAL2(4);
		PDF_VAL(8) = PDF_VAL2(8);
		PDF_VAL(9) = PDF_VAL2(9);
        }*/ 
}

/**
 * Kernel - Apply outer boundaries
 * Second variant, only applies boundaries, not run for the whole domain
 */
__global__ void applyOuterBoundariesGPU_kernelYZ(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ, float3 inVel, float inRho)
{
        int x;
        int y = threadIdx.x;
        int z;

        #if PRINT_DEBUG_KERNEL
                if(x == 0 && y == 0 && z == 0)
                        printf("Inflow boundary kernel: Received invel %f, %f, %f\n", 
				inVel.x, inVel.y, inVel.z);
        #endif
       	if(blockIdx.y < dimZ) {
		x = blockIdx.x;
		z = blockIdx.y;
	}
	else {
		x = dimX - *lbm_boundary_size_dev + blockIdx.x;
		z = blockIdx.y - dimZ;
	}
	SET_PDFS_TO_VEL_PRECISE(pdf, x, y, z, inRho, inVel.x, inVel.y, inVel.z)
}

/**
 * Kernel - Apply outer boundaries
 * Second variant, only applies boundaries, not run for the whole domain
 */
__global__ void applyOuterBoundariesGPU_kernelXY(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ, float3 inVel, float inRho)
{
        int x = threadIdx.x;
        int y;
        int z;

        #if PRINT_DEBUG_KERNEL
                if(x == 0 && y == 0 && z == 0)
                        printf("Inflow boundary kernel: Received invel %f, %f, %f\n", 
				inVel.x, inVel.y, inVel.z);
        #endif
	
	if(blockIdx.x < dimY) {
		z = blockIdx.y;
		y = blockIdx.x;
	}
	else {
		z = dimZ - *lbm_boundary_size_dev + blockIdx.y;
		y = blockIdx.x - dimY;
	}
        	
	SET_PDFS_TO_VEL_PRECISE(pdf, x, y, z, inRho, inVel.x, inVel.y, inVel.z)
}

/**
 * Kernel - Apply outer boundaries
 * Second variant, only applies boundaries, not run for the whole domain
 */
__global__ void applyOuterBoundariesGPU_kernelXZ(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ, float3 inVel, float inRho)
{
        int x = threadIdx.x;
        int y;
        int z;

        #if PRINT_DEBUG_KERNEL
                if(x == 0 && y == 0 && z == 0)
                        printf("Inflow boundary kernel: Received invel %f, %f, %f\n", 
				inVel.x, inVel.y, inVel.z);
        #endif
	
	if(blockIdx.x < dimZ) {
		y = blockIdx.y;
		z = blockIdx.x;
	}
	else {
		y = dimY - *lbm_boundary_size_dev + blockIdx.y;
		z = blockIdx.x - dimZ;
	}
        	
	SET_PDFS_TO_VEL_PRECISE(pdf, x, y, z, inRho, inVel.x, inVel.y, inVel.z)
}

/**
 * Kernel - Apply outer boundaries
 * Second variant, only applies boundaries, not run for the whole domain
 */
__global__ void applyOutFlowGPU_kernel(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ)
{
        int x = dimX - *lbm_boundary_size_dev + blockIdx.y;
        int y = threadIdx.x;
        int z = blockIdx.x;

        #if PRINT_DEBUG_KERNEL
                if(x == 0 && y == 0 && z == 0)
                        printf("Outflow boundary kernel: Received invel %f, %f, %f\n", 
				inVel.x, inVel.y, inVel.z);
        #endif
	
        #define PDF_VAL2(n) (*((float_type*) ((char*)pdf.ptr[n] + offset2)))               
        volatile int offset = (int) &PDF_REF(float_type, 0, x, y, z, *lbm_pitch_dev, *lbm_ysize_dev);
	// Offset set to the last Y/Z-plane before the boundary
        volatile int offset2 = (int) &PDF_REF(float_type, 0, dimX - *lbm_boundary_size_dev - 1, y, z, *lbm_pitch_dev, *lbm_ysize_dev);
 	 
	PDF_VAL(14) = PDF_VAL2(14);
	PDF_VAL(2) = PDF_VAL2(2);
	PDF_VAL(4) = PDF_VAL2(4);
	PDF_VAL(8) = PDF_VAL2(8);
	PDF_VAL(9) = PDF_VAL2(9);
}

/**
 * Kernel - Apply initial PDF values
 */
__global__ void applyInitialValuesGPU_kernel(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ, float xInitValue, float rhoInitValue)
{
        int x = threadIdx.x;
        int y = blockIdx.x / solid_d.ysize;
        int z = blockIdx.x - __mul24(y, solid_d.ysize);

        int mask = OBS_REF(int, solid_d, x, y, z);
        SET_PDFS_TO_VEL_PRECISE(pdf, x, y, z, rhoInitValue, xInitValue, 0.0, 0.0)
}

/**
 * Kernel - Reset PDF values to zero
 */
__global__ void applyValueToAllGPU_kernel(cudaPitchedPtr solid_d, PDF pdf, int f_pitch, size_t ysize, int dimX, int dimY, int dimZ)
{
        int x = threadIdx.x;
        int y = blockIdx.x / solid_d.ysize;
        int z = blockIdx.x - __mul24(y, solid_d.ysize);

        int mask = OBS_REF(int, solid_d, x, y, z);

        volatile int offset = (int) &PDF_REF(float_type, 0, x, y, z, f_pitch, ysize);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

	for(int i=0;i<19;i++) {
		PDF_VAL(i) = 0.0;
	}
}

/**
 * Kernel - Get Next X
 */
__device__ void getNextX(int x, int size_x, int f_index, bool periodic, int &n_x)
{
	int nextX = x + c_3d_dev[f_index][0];

	if(periodic)
	{
		if(nextX == -1)
			nextX = size_x-1;

		if(nextX == size_x)
			nextX = 0;
	}
	
	n_x = nextX;
}

/**
 * Kernel - Get Next Y
 */
__device__ void getNextY(int y, int size_y, int f_index, bool periodic, int &n_y)
{
	int nextY = y + c_3d_dev[f_index][1];

	if(periodic)
	{
		if(nextY == -1)
			nextY = size_y-1;

		if(nextY == size_y)
			nextY = 0;
	}

	n_y = nextY;
}

/**
 * Kernel - Get Next Z
 */
__device__ void getNextZ(int z, int size_z, int f_index, bool periodic, int &n_z)
{
	int nextZ = z + c_3d_dev[f_index][2];

	if(periodic)
	{
		if(nextZ == -1)
			nextZ = size_z-1;

		if(nextZ == size_z)	
			nextZ = 0;
	}

	n_z = nextZ;
} 

/**
 * Kernel - Stream And Swap
 */
__global__ void streamAndSwapGPU_kernel(cudaPitchedPtr solid_d, PDF pdf, size_t f_pitch, int width, int height, int depth, int extra)
{	
	/* Old indexing scheme, for large domain sizes
	int x = threadIdx.x;
	int grid_y = __mul24(blockIdx.y, gridDim.x);
	int z = (blockIdx.x + grid_y) / height;
	int y = (blockIdx.x + grid_y) - z * height;*/

    int x = threadIdx.x;
    int y = blockIdx.x;
    int z = blockIdx.y; 

    if (x < width && y < height && z < depth) {
                //if(x % 40 == 0)
                //        printf("Stream and swap kernel running on pos: %d, %d, %d\n", x, y, z);
	//if (x < width && y < height && z < depth && x >= *lbm_boundary_size_dev && x < (width - *lbm_boundary_size_dev)) {	
		int nextX, nextY, nextZ;		

		//uchar solid = OBS(solid_d, x, y, z);	
				
		//if(solid == 0)
		if(OBS_REF(int, solid_d, x, y, z) == 0)
		{
			bool periodicX;
			bool periodicY;
			bool periodicZ;
			/*
			if((x >= (extra/2)) && (x <= width-(extra/2)))
			{
			    //periodicX = true;
				//periodicY = false;
				//periodicZ = false;

				periodicX = true;
				periodicY = true;
				periodicZ = true;
			}
			else 
			{
			    periodicX = true;
				periodicY = true;
				periodicZ = true;
			}
			*/
			periodicX = true;
			periodicY = true;
			periodicZ = true;
			
			float_type swap_tmp;
			float_type swap_tmp2;
/*
			//Swap 1 and 10, 2 and 11, etc... ... 9 and 18
			for(int i=1; i<10; i++)
			{
				getNextX(x, width, i, periodicX, nextX);	
				getNextY(y, height, i, periodicY, nextY);		
				getNextZ(z, depth, i, periodicZ, nextZ);

				if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
				{
					if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
					{	
						swap_tmp = PDF_REF(float_type, pdf.ptr[i+9], x, y, z, f_pitch, height);
						swap_tmp2 = PDF_REF(float_type, pdf.ptr[i], nextX, nextY, nextZ, f_pitch, height);
						PDF_REF(float_type, pdf.ptr[i+9], x, y, z, f_pitch, height) = swap_tmp2;
						PDF_REF(float_type, pdf.ptr[i], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
					}
				}
			}
*/

			
			// Swap 1 and 10s
			
			getNextX(x, width, 1, periodicX, nextX);	
			getNextY(y, height, 1, periodicY, nextY);		
			getNextZ(z, depth, 1, periodicZ, nextZ);
		
			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{ 	
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[10], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[1], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[10], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[1], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}
	
			// Swap 2 and 11
			
			getNextX(x, width, 2, periodicX, nextX);	
			getNextY(y, height, 2, periodicY, nextY);		
			getNextZ(z, depth, 2, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[11], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[2], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[11], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[2], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 3 and 12
			
			getNextX(x, width, 3, periodicX, nextX);	
			getNextY(y, height, 3, periodicY, nextY);		
			getNextZ(z, depth, 3, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[12], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[3], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[12], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[3], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 4 and 13
			
			getNextX(x, width, 4, periodicX, nextX);	
			getNextY(y, height, 4, periodicY, nextY);		
			getNextZ(z, depth, 4, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[13], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[4], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[13], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[4], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 5 and 14
			
			getNextX(x, width, 5, periodicX, nextX);	
			getNextY(y, height, 5, periodicY, nextY);		
			getNextZ(z, depth, 5, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[14], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[5], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[14], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[5], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 6 and 15
			
			getNextX(x, width, 6, periodicX, nextX);	
			getNextY(y, height, 6, periodicY, nextY);		
			getNextZ(z, depth, 6, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[15], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[6], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[15], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[6], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 7 and 16
			
			getNextX(x, width, 7, periodicX, nextX);	
			getNextY(y, height, 7, periodicY, nextY);		
			getNextZ(z, depth, 7, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[16], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[7], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[16], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[7], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 8 and 17
			
			getNextX(x, width, 8, periodicX, nextX);	
			getNextY(y, height, 8, periodicY, nextY);		
			getNextZ(z, depth, 8, periodicZ, nextZ);
			
			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[17], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[8], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[17], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[8], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}

			// Swap 9 and 18
			
			getNextX(x, width, 9, periodicX, nextX);	
			getNextY(y, height, 9, periodicY, nextY);		
			getNextZ(z, depth, 9, periodicZ, nextZ);

			if(nextX >= 0 && nextY >= 0 && nextZ >= 0 && nextX < width && nextY < height && nextZ < depth)
			{
				if(OBS_REF(int, solid_d, nextX, nextY, nextZ) == 0)
				{	
					//float_type swap_tmp;
					swap_tmp = PDF_REF(float_type, pdf.ptr[18], x, y, z, f_pitch, height);
					swap_tmp2 = PDF_REF(float_type, pdf.ptr[9], nextX, nextY, nextZ, f_pitch, height);
					PDF_REF(float_type, pdf.ptr[18], x, y, z, f_pitch, height) = swap_tmp2;
					PDF_REF(float_type, pdf.ptr[9], nextX, nextY, nextZ, f_pitch, height) = swap_tmp;
				}
			}
			
		}
	}
}

extern __shared__ float_type shared_lbm[];

/**
 * Utility function to get the value of rho from the PDFs (shared memory)
 */
__device__ float_type getRhoShared(int share_x)
{
	return (shared_lbm[share_x + 4] + shared_lbm[share_x + 5] +  
	        shared_lbm[share_x + 6] + shared_lbm[share_x + 7] +  
	        shared_lbm[share_x + 8] + shared_lbm[share_x + 9] +  
	        shared_lbm[share_x + 13] + shared_lbm[share_x + 14] +   
	        shared_lbm[share_x + 15] + shared_lbm[share_x + 16] + 
	        shared_lbm[share_x + 17] + shared_lbm[share_x + 18] + 
	        shared_lbm[share_x + 1] +  shared_lbm[share_x + 2] + 
	        shared_lbm[share_x + 3] + shared_lbm[share_x + 11] +  
	        shared_lbm[share_x + 12] + shared_lbm[share_x + 10] + 
	        shared_lbm[share_x + 0]); 
}

/**
 * Utility function to get the x-component of the velocity from the PDFs (shared memory)
 */
__device__ float_type getUxShared(int share_x, float_type rho)
{
	return ((shared_lbm[share_x + 13] - shared_lbm[share_x + 4]) + 
	        (shared_lbm[share_x + 14] - shared_lbm[share_x + 5]) + 
	        (shared_lbm[share_x + 15] - shared_lbm[share_x + 6]) + 
	        (shared_lbm[share_x + 16] - shared_lbm[share_x + 7]) + 
	        (shared_lbm[share_x + 10] - shared_lbm[share_x + 1])) / rho;
}

/**
 * Utility function to get the y-component of the velocity from the PDFs (shared memory)
 */
__device__ float_type getUyShared(int share_x, float_type rho)
{
   	return ((shared_lbm[share_x + 5] - shared_lbm[share_x + 14]) +  
   	        (shared_lbm[share_x + 11] - shared_lbm[share_x + 2]) +  
   	        (shared_lbm[share_x + 13] - shared_lbm[share_x + 4]) + 
   	        (shared_lbm[share_x + 17] - shared_lbm[share_x + 8]) + 
   	        (shared_lbm[share_x + 18] - shared_lbm[share_x + 9])) / rho;
}		

/**
 * Utility function to get the z-component of the velocity from the PDFs (shared memory)
 */
__device__ float_type getUzShared(int share_x, float_type rho)
{
	 return ((shared_lbm[share_x + 7] - shared_lbm[share_x + 16]) +  
	         (shared_lbm[share_x + 9] - shared_lbm[share_x + 18]) + 
	         (shared_lbm[share_x + 15] - shared_lbm[share_x + 6]) +  
	         (shared_lbm[share_x + 17] - shared_lbm[share_x + 8]) + 
	         (shared_lbm[share_x + 12] - shared_lbm[share_x + 3])) / rho;			
}		

/**
 * Kernel - Collide And Swap 3
 */
__global__ void collideAndSwapGPU_kernelShared3(float_type omega, float_type force, cudaPitchedPtr solid_d, PDF pdf, size_t pitch, size_t ysize, int width, int height, int depth)
{
	/*
	int y = threadIdx.x;
    int x = blockIdx.x;
    int z = blockIdx.y;
	/*/
    // Orig
    //*
	int x = threadIdx.x;
	int grid_y = __mul24(blockIdx.y, gridDim.x);
	int z = (blockIdx.x + grid_y) / height;
	int y = (blockIdx.x + grid_y) - z * height;
	//*/

        /*int x = blockIdx.x;
        int grid_y = __mul24(blockIdx.y, blockDim.x);
        int z = (threadIdx.x + grid_y) / blockDim.x;
        int y = (threadIdx.x + grid_y) - z * blockDim.x;*/
	if (x < width && y < height && z < depth)
	{
		if(OBS_REF(int, solid_d, x, y, z) == 0)
		{
	volatile int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, ysize);
	
	#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
	
	volatile int share_x = y*19;
	
	shared_lbm[share_x + 0] = PDF_VAL(0);
	shared_lbm[share_x + 1] = PDF_VAL(1);
	shared_lbm[share_x + 2] = PDF_VAL(2);
	shared_lbm[share_x + 3] = PDF_VAL(3);
	shared_lbm[share_x + 4] = PDF_VAL(4);
	shared_lbm[share_x + 5] = PDF_VAL(5);
	shared_lbm[share_x + 6] = PDF_VAL(6);
	shared_lbm[share_x + 7] = PDF_VAL(7);
	shared_lbm[share_x + 8] = PDF_VAL(8);
	shared_lbm[share_x + 9] = PDF_VAL(9);
	shared_lbm[share_x + 10] = PDF_VAL(10);
	shared_lbm[share_x + 11] = PDF_VAL(11);
	shared_lbm[share_x + 12] = PDF_VAL(12);
	shared_lbm[share_x + 13] = PDF_VAL(13);
	shared_lbm[share_x + 14] = PDF_VAL(14);
	shared_lbm[share_x + 15] = PDF_VAL(15);
	shared_lbm[share_x + 16] = PDF_VAL(16);
	shared_lbm[share_x + 17] = PDF_VAL(17);
	shared_lbm[share_x + 18] = PDF_VAL(18);
	
	__syncthreads();

	
			float_type cu;	
			
			float_type omega_a = (1.f - omega);
	
			// Compute density from the particle distribution functions 
			float_type rho;
			rho = getRhoShared(share_x);
					 	
			// Compute velocity in x, y, and z directions from the particle distribution functions
			float_type ux; 
			ux = getUxShared(share_x, rho);
			
			float_type uy; 
			uy =  getUyShared(share_x, rho);
		
			float_type uz;
			uz =  getUzShared(share_x, rho);
	
			float_type uSqr = ux*ux + uy*uy + uz*uz;
			
			float_type u_sqr = 1.5f * uSqr;
				
			float_type a = 4.5f;
			
			// 0
			cu = c_3d_dev[0][0]*ux + c_3d_dev[0][1]*uy + c_3d_dev[0][2]*uz;
			float_type f0_temp = t_3d_dev[0] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f0_temp *= rho;
			
			// 1
                        cu = c_3d_dev[1][0]*ux + c_3d_dev[1][1]*uy + c_3d_dev[1][2]*uz;
			float_type f1_temp = t_3d_dev[1] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f1_temp *= rho;
			
			// 2
			cu = c_3d_dev[2][0]*ux + c_3d_dev[2][1]*uy + c_3d_dev[2][2]*uz;
			float_type f2_temp = t_3d_dev[2] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f2_temp *= rho;
			
				// 3
			cu = c_3d_dev[3][0]*ux + c_3d_dev[3][1]*uy + c_3d_dev[3][2]*uz;
			float_type f3_temp = t_3d_dev[3] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f3_temp *= rho;
			
			// 4
			cu = c_3d_dev[4][0]*ux + c_3d_dev[4][1]*uy + c_3d_dev[4][2]*uz;
			float_type f4_temp = t_3d_dev[4] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f4_temp *= rho;
			
			// 5
			cu = c_3d_dev[5][0]*ux + c_3d_dev[5][1]*uy + c_3d_dev[5][2]*uz;
			float_type f5_temp = t_3d_dev[5] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f5_temp *= rho;
			
			// 6
			cu = c_3d_dev[6][0]*ux + c_3d_dev[6][1]*uy + c_3d_dev[6][2]*uz;
			float_type f6_temp = t_3d_dev[6] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);		
			f6_temp *= rho;
			
			// 7
			cu = c_3d_dev[7][0]*ux + c_3d_dev[7][1]*uy + c_3d_dev[7][2]*uz;
			float_type f7_temp = t_3d_dev[7] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f7_temp *= rho;
			
			// 8
			cu = c_3d_dev[8][0]*ux + c_3d_dev[8][1]*uy + c_3d_dev[8][2]*uz;
			float_type f8_temp = t_3d_dev[8] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f8_temp *= rho;
			
			// 9
			cu = c_3d_dev[9][0]*ux + c_3d_dev[9][1]*uy + c_3d_dev[9][2]*uz;
			float_type f9_temp = t_3d_dev[9] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f9_temp *= rho;
		
			// 10
			cu = c_3d_dev[10][0]*ux + c_3d_dev[10][1]*uy + c_3d_dev[10][2]*uz;
			float_type f10_temp = t_3d_dev[10] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f10_temp *= rho;
			
			// 11 
		        cu = c_3d_dev[11][0]*ux + c_3d_dev[11][1]*uy + c_3d_dev[11][2]*uz;
			float_type f11_temp = t_3d_dev[11] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f11_temp *= rho;
		
			// 12
			cu = c_3d_dev[12][0]*ux + c_3d_dev[12][1]*uy + c_3d_dev[12][2]*uz;
			float_type f12_temp = t_3d_dev[12] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f12_temp *= rho;

			// 13
			cu = c_3d_dev[13][0]*ux + c_3d_dev[13][1]*uy + c_3d_dev[13][2]*uz;
			float_type f13_temp = t_3d_dev[13] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f13_temp *= rho;

			// 14
			cu = c_3d_dev[14][0]*ux + c_3d_dev[14][1]*uy + c_3d_dev[14][2]*uz;
			float_type f14_temp = t_3d_dev[14] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f14_temp *= rho;
			
			// 15
			cu = c_3d_dev[15][0]*ux + c_3d_dev[15][1]*uy + c_3d_dev[15][2]*uz;
			float_type f15_temp = t_3d_dev[15] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f15_temp *= rho;
			
			// 16
			cu = c_3d_dev[16][0]*ux + c_3d_dev[16][1]*uy + c_3d_dev[16][2]*uz;
			float_type f16_temp = t_3d_dev[16] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f16_temp *= rho;
		
			// 17
			cu = c_3d_dev[17][0]*ux + c_3d_dev[17][1]*uy + c_3d_dev[17][2]*uz;
			float_type f17_temp = t_3d_dev[17] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f17_temp *= rho;
			
			// 18
			cu = c_3d_dev[18][0]*ux + c_3d_dev[18][1]*uy + c_3d_dev[18][2]*uz;
			float_type f18_temp = t_3d_dev[18] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f18_temp *= rho;
			
			float_type rho2 = rho * 0.333333333333f;
			
			float_type tmp0 = shared_lbm[share_x + 0];
			tmp0 += ((f0_temp - tmp0)*omega) + 
			                       ((c_3d_dev[0][0] - ux) * f0_temp ) / (rho2) * force;
			PDF_VAL(0) = tmp0;
						
			float_type tmp1 = shared_lbm[share_x + 1];
			tmp1 += ((f1_temp - tmp1)*omega) + 
			                       ((c_3d_dev[1][0] - ux) * f1_temp ) / (rho2) * force;
			PDF_VAL(10) = tmp1;	
			
			float_type tmp10 = shared_lbm[share_x + 10];
			tmp10 += ((f10_temp - tmp10)*omega) + 
			                        ((c_3d_dev[10][0] - ux) * f10_temp ) / (rho2) * force;
			PDF_VAL(1) = tmp10;
			
			float_type tmp2 = shared_lbm[share_x + 2];
			tmp2 += ((f2_temp - tmp2)*omega) + 
			                       ((c_3d_dev[2][0] - ux) * f2_temp ) / (rho2) * force;
			PDF_VAL(11) = tmp2;
		
			float_type tmp11 = shared_lbm[share_x + 11];
			tmp11 += ((f11_temp - tmp11)*omega) + 
			                        ((c_3d_dev[11][0] - ux) * f11_temp ) / (rho2) * force;
			PDF_VAL(2) = tmp11;

			float_type tmp3 = shared_lbm[share_x + 3];
			tmp3 += ((f3_temp - tmp3)*omega) + 
			                        ((c_3d_dev[3][0] - ux) * f3_temp ) / (rho2) * force;
			PDF_VAL(12) = tmp3;

			float_type tmp12 = shared_lbm[share_x + 12];
			tmp12 += ((f12_temp - tmp12)*omega) + 
			                        ((c_3d_dev[12][0] - ux) * f12_temp ) / (rho2) * force;
			PDF_VAL(3) = tmp12;
			
			float_type tmp4 = shared_lbm[share_x + 4];
			tmp4 += ((f4_temp - tmp4)*omega) + 
			                       ((c_3d_dev[4][0] - ux) * f4_temp ) / (rho2) * force;
			PDF_VAL(13) = tmp4;
			
			float_type tmp13 = shared_lbm[share_x + 13];
			tmp13 += ((f13_temp - tmp13)*omega) + 
			                        ((c_3d_dev[13][0] - ux) * f13_temp ) / (rho2) * force;
			PDF_VAL(4) = tmp13;
			
			float_type tmp5 = shared_lbm[share_x + 5];
			tmp5 += ((f5_temp - tmp5)*omega) + 
			                        ((c_3d_dev[5][0] - ux) * f5_temp ) / (rho2) * force;
			PDF_VAL(14) = tmp5;
	
			float_type tmp14 = shared_lbm[share_x + 14];
			tmp14 += ((f14_temp - tmp14)*omega) + 
			                        ((c_3d_dev[14][0] - ux) * f14_temp ) / (rho2) * force;
			PDF_VAL(5) = tmp14;
			
			float_type tmp6 = shared_lbm[share_x + 6];
		        tmp6 += ((f6_temp - tmp6)*omega) + 
			                       ((c_3d_dev[6][0] - ux) * f6_temp ) / (rho2) * force;
			PDF_VAL(15) = tmp6;
			
			float_type tmp15 = shared_lbm[share_x + 15];
		        tmp15 += ((f15_temp - tmp15)*omega) + 
			                        ((c_3d_dev[15][0] - ux) * f15_temp ) / (rho2) * force;
			PDF_VAL(6) = tmp15;
			
			float_type tmp7 = shared_lbm[share_x + 7];
		        tmp7 += ((f7_temp - tmp7)*omega) + 
			                       ((c_3d_dev[7][0] - ux) * f7_temp ) / (rho2) * force;
			PDF_VAL(16) = tmp7;
			
			float_type tmp16 = shared_lbm[share_x + 16];
		        tmp16 += ((f16_temp - tmp16)*omega) + 
			                        ((c_3d_dev[16][0] - ux) * f16_temp ) / (rho2) * force;
			PDF_VAL(7) = tmp16;
			
			float_type tmp8 = shared_lbm[share_x + 8];
		        tmp8 += ((f8_temp - tmp8)*omega) + 
			                        ((c_3d_dev[8][0] - ux) * f8_temp ) / (rho2) * force;
			PDF_VAL(17) = tmp8;

			float_type tmp17 = shared_lbm[share_x + 17];
		        tmp17 += ((f17_temp - tmp17)*omega) + 
			                        ((c_3d_dev[17][0] - ux) * f17_temp ) / (rho2) * force;
			PDF_VAL(8) = tmp17;
			
			float_type tmp9 = shared_lbm[share_x + 9];
		        tmp9 += ((f9_temp - tmp9)*omega) + 
			                        ((c_3d_dev[9][0] - ux) * f9_temp ) / (rho2) * force;
			PDF_VAL(18) = tmp9;

			float_type tmp18 = shared_lbm[share_x + 18];
		        tmp18 += ((f18_temp - tmp18)*omega) + 
			                        ((c_3d_dev[18][0] - ux) * f18_temp ) / (rho2) * force;
			PDF_VAL(9) = tmp18;
		}
	}
}

/**
 * Kernel - Collide And Swap 3.2
 * Shared memory based collide kernel for Fermi
 */
#define SHAREDPDFSIZE 16
__global__ void collideAndSwapGPU_kernelShared4(float_type omega, float_type force, cudaPitchedPtr solid_d, PDF pdf, size_t pitch, size_t ysize, int width, int height, int depth)
{
	int y = threadIdx.x;
    int x = blockIdx.x;
    int z = blockIdx.y;
	
	if (x < width && y < height && z < depth)
	{
		if(OBS_REF(int, solid_d, x, y, z) == 0)
		{
			volatile int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, ysize);
			#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))

			__shared__ float_type shared_lbm2[SHAREDPDFSIZE * 19];
			volatile int share_x = y*19;
			
			shared_lbm2[share_x + 0] = PDF_VAL(0);
			shared_lbm2[share_x + 1] = PDF_VAL(1);
			shared_lbm2[share_x + 2] = PDF_VAL(2);
			shared_lbm2[share_x + 3] = PDF_VAL(3);
			shared_lbm2[share_x + 4] = PDF_VAL(4);
			shared_lbm2[share_x + 5] = PDF_VAL(5);
			shared_lbm2[share_x + 6] = PDF_VAL(6);
			shared_lbm2[share_x + 7] = PDF_VAL(7);
			shared_lbm2[share_x + 8] = PDF_VAL(8);
			shared_lbm2[share_x + 9] = PDF_VAL(9);
			shared_lbm2[share_x + 10] = PDF_VAL(10);
			shared_lbm2[share_x + 11] = PDF_VAL(11);
			shared_lbm2[share_x + 12] = PDF_VAL(12);
			shared_lbm2[share_x + 13] = PDF_VAL(13);
			shared_lbm2[share_x + 14] = PDF_VAL(14);
			shared_lbm2[share_x + 15] = PDF_VAL(15);
			shared_lbm2[share_x + 16] = PDF_VAL(16);
			shared_lbm2[share_x + 17] = PDF_VAL(17);
			shared_lbm2[share_x + 18] = PDF_VAL(18);
			
			__syncthreads();

			float_type cu;
			float_type omega_a = (1.f - omega);
	
			// Compute density from the particle distribution functions 
			float_type rho=(shared_lbm2[share_x + 4] + shared_lbm2[share_x + 5] +  
							shared_lbm2[share_x + 6] + shared_lbm2[share_x + 7] +  
							shared_lbm2[share_x + 8] + shared_lbm2[share_x + 9] +  
							shared_lbm2[share_x + 13] + shared_lbm2[share_x + 14] +   
							shared_lbm2[share_x + 15] + shared_lbm2[share_x + 16] + 
							shared_lbm2[share_x + 17] + shared_lbm2[share_x + 18] + 
							shared_lbm2[share_x + 1] +  shared_lbm2[share_x + 2] + 
							shared_lbm2[share_x + 3] + shared_lbm2[share_x + 11] +  
							shared_lbm2[share_x + 12] + shared_lbm2[share_x + 10] + 
							shared_lbm2[share_x + 0]);

			// Compute velocity in x, y, and z directions from the particle distribution functions
			float_type ux= ((shared_lbm2[share_x + 13] - shared_lbm2[share_x + 4]) + 
							(shared_lbm2[share_x + 14] - shared_lbm2[share_x + 5]) + 
							(shared_lbm2[share_x + 15] - shared_lbm2[share_x + 6]) + 
							(shared_lbm2[share_x + 16] - shared_lbm2[share_x + 7]) + 
							(shared_lbm2[share_x + 10] - shared_lbm2[share_x + 1]))
							/ rho;
			
			float_type uy= ((shared_lbm2[share_x + 5] - shared_lbm2[share_x + 14]) +  
   							(shared_lbm2[share_x + 11] - shared_lbm2[share_x + 2]) +  
   							(shared_lbm2[share_x + 13] - shared_lbm2[share_x + 4]) + 
   							(shared_lbm2[share_x + 17] - shared_lbm2[share_x + 8]) + 
   							(shared_lbm2[share_x + 18] - shared_lbm2[share_x + 9]))
							/ rho;

			float_type uz = ((shared_lbm2[share_x + 7] - shared_lbm2[share_x + 16]) +  
							 (shared_lbm2[share_x + 9] - shared_lbm2[share_x + 18]) + 
							 (shared_lbm2[share_x + 15] - shared_lbm2[share_x + 6]) +  
							 (shared_lbm2[share_x + 17] - shared_lbm2[share_x + 8]) + 
							 (shared_lbm2[share_x + 12] - shared_lbm2[share_x + 3]))
							 / rho;

			float_type uSqr = ux*ux + uy*uy + uz*uz;
			float_type u_sqr = 1.5f * uSqr;
			float_type a = 4.5f;

			// 0
			cu = c_3d_dev[0][0]*ux + c_3d_dev[0][1]*uy + c_3d_dev[0][2]*uz;
			float_type f0_temp = t_3d_dev[0] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f0_temp *= rho;
			
			// 1
            cu = c_3d_dev[1][0]*ux + c_3d_dev[1][1]*uy + c_3d_dev[1][2]*uz;
			float_type f1_temp = t_3d_dev[1] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f1_temp *= rho;
			
			// 2
			cu = c_3d_dev[2][0]*ux + c_3d_dev[2][1]*uy + c_3d_dev[2][2]*uz;
			float_type f2_temp = t_3d_dev[2] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f2_temp *= rho;
			
				// 3
			cu = c_3d_dev[3][0]*ux + c_3d_dev[3][1]*uy + c_3d_dev[3][2]*uz;
			float_type f3_temp = t_3d_dev[3] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f3_temp *= rho;
			
			// 4
			cu = c_3d_dev[4][0]*ux + c_3d_dev[4][1]*uy + c_3d_dev[4][2]*uz;
			float_type f4_temp = t_3d_dev[4] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f4_temp *= rho;
			
			// 5
			cu = c_3d_dev[5][0]*ux + c_3d_dev[5][1]*uy + c_3d_dev[5][2]*uz;
			float_type f5_temp = t_3d_dev[5] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f5_temp *= rho;
			
			// 6
			cu = c_3d_dev[6][0]*ux + c_3d_dev[6][1]*uy + c_3d_dev[6][2]*uz;
			float_type f6_temp = t_3d_dev[6] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);		
			f6_temp *= rho;
			
			// 7
			cu = c_3d_dev[7][0]*ux + c_3d_dev[7][1]*uy + c_3d_dev[7][2]*uz;
			float_type f7_temp = t_3d_dev[7] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f7_temp *= rho;
			
			// 8
			cu = c_3d_dev[8][0]*ux + c_3d_dev[8][1]*uy + c_3d_dev[8][2]*uz;
			float_type f8_temp = t_3d_dev[8] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f8_temp *= rho;
			
			// 9
			cu = c_3d_dev[9][0]*ux + c_3d_dev[9][1]*uy + c_3d_dev[9][2]*uz;
			float_type f9_temp = t_3d_dev[9] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f9_temp *= rho;
		
			// 10
			cu = c_3d_dev[10][0]*ux + c_3d_dev[10][1]*uy + c_3d_dev[10][2]*uz;
			float_type f10_temp = t_3d_dev[10] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f10_temp *= rho;
			
			// 11 
		        cu = c_3d_dev[11][0]*ux + c_3d_dev[11][1]*uy + c_3d_dev[11][2]*uz;
			float_type f11_temp = t_3d_dev[11] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f11_temp *= rho;
		
			// 12
			cu = c_3d_dev[12][0]*ux + c_3d_dev[12][1]*uy + c_3d_dev[12][2]*uz;
			float_type f12_temp = t_3d_dev[12] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f12_temp *= rho;

			// 13
			cu = c_3d_dev[13][0]*ux + c_3d_dev[13][1]*uy + c_3d_dev[13][2]*uz;
			float_type f13_temp = t_3d_dev[13] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f13_temp *= rho;

			// 14
			cu = c_3d_dev[14][0]*ux + c_3d_dev[14][1]*uy + c_3d_dev[14][2]*uz;
			float_type f14_temp = t_3d_dev[14] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f14_temp *= rho;
			
			// 15
			cu = c_3d_dev[15][0]*ux + c_3d_dev[15][1]*uy + c_3d_dev[15][2]*uz;
			float_type f15_temp = t_3d_dev[15] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f15_temp *= rho;
			
			// 16
			cu = c_3d_dev[16][0]*ux + c_3d_dev[16][1]*uy + c_3d_dev[16][2]*uz;
			float_type f16_temp = t_3d_dev[16] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			f16_temp *= rho;
		
			// 17
			cu = c_3d_dev[17][0]*ux + c_3d_dev[17][1]*uy + c_3d_dev[17][2]*uz;
			float_type f17_temp = t_3d_dev[17] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f17_temp *= rho;
			
			// 18
			cu = c_3d_dev[18][0]*ux + c_3d_dev[18][1]*uy + c_3d_dev[18][2]*uz;
			float_type f18_temp = t_3d_dev[18] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			f18_temp *= rho;
			
			float_type rho2 = rho * 0.333333333333f;
			
			float_type tmp0 = shared_lbm2[share_x + 0];
			tmp0 += ((f0_temp - tmp0)*omega) + 
			                       ((c_3d_dev[0][0] - ux) * f0_temp ) / (rho2) * force;
			PDF_VAL(0) = tmp0;
						
			float_type tmp1 = shared_lbm2[share_x + 1];
			tmp1 += ((f1_temp - tmp1)*omega) + 
			                       ((c_3d_dev[1][0] - ux) * f1_temp ) / (rho2) * force;
			PDF_VAL(10) = tmp1;	
			
			float_type tmp10 = shared_lbm2[share_x + 10];
			tmp10 += ((f10_temp - tmp10)*omega) + 
			                        ((c_3d_dev[10][0] - ux) * f10_temp ) / (rho2) * force;
			PDF_VAL(1) = tmp10;
			
			float_type tmp2 = shared_lbm2[share_x + 2];
			tmp2 += ((f2_temp - tmp2)*omega) + 
			                       ((c_3d_dev[2][0] - ux) * f2_temp ) / (rho2) * force;
			PDF_VAL(11) = tmp2;
		
			float_type tmp11 = shared_lbm2[share_x + 11];
			tmp11 += ((f11_temp - tmp11)*omega) + 
			                        ((c_3d_dev[11][0] - ux) * f11_temp ) / (rho2) * force;
			PDF_VAL(2) = tmp11;

			float_type tmp3 = shared_lbm2[share_x + 3];
			tmp3 += ((f3_temp - tmp3)*omega) + 
			                        ((c_3d_dev[3][0] - ux) * f3_temp ) / (rho2) * force;
			PDF_VAL(12) = tmp3;

			float_type tmp12 = shared_lbm2[share_x + 12];
			tmp12 += ((f12_temp - tmp12)*omega) + 
			                        ((c_3d_dev[12][0] - ux) * f12_temp ) / (rho2) * force;
			PDF_VAL(3) = tmp12;
			
			float_type tmp4 = shared_lbm2[share_x + 4];
			tmp4 += ((f4_temp - tmp4)*omega) + 
			                       ((c_3d_dev[4][0] - ux) * f4_temp ) / (rho2) * force;
			PDF_VAL(13) = tmp4;
			
			float_type tmp13 = shared_lbm2[share_x + 13];
			tmp13 += ((f13_temp - tmp13)*omega) + 
			                        ((c_3d_dev[13][0] - ux) * f13_temp ) / (rho2) * force;
			PDF_VAL(4) = tmp13;
			
			float_type tmp5 = shared_lbm2[share_x + 5];
			tmp5 += ((f5_temp - tmp5)*omega) + 
			                        ((c_3d_dev[5][0] - ux) * f5_temp ) / (rho2) * force;
			PDF_VAL(14) = tmp5;
	
			float_type tmp14 = shared_lbm2[share_x + 14];
			tmp14 += ((f14_temp - tmp14)*omega) + 
			                        ((c_3d_dev[14][0] - ux) * f14_temp ) / (rho2) * force;
			PDF_VAL(5) = tmp14;
			
			float_type tmp6 = shared_lbm2[share_x + 6];
		        tmp6 += ((f6_temp - tmp6)*omega) + 
			                       ((c_3d_dev[6][0] - ux) * f6_temp ) / (rho2) * force;
			PDF_VAL(15) = tmp6;
			
			float_type tmp15 = shared_lbm2[share_x + 15];
		        tmp15 += ((f15_temp - tmp15)*omega) + 
			                        ((c_3d_dev[15][0] - ux) * f15_temp ) / (rho2) * force;
			PDF_VAL(6) = tmp15;
			
			float_type tmp7 = shared_lbm2[share_x + 7];
		        tmp7 += ((f7_temp - tmp7)*omega) + 
			                       ((c_3d_dev[7][0] - ux) * f7_temp ) / (rho2) * force;
			PDF_VAL(16) = tmp7;
			
			float_type tmp16 = shared_lbm2[share_x + 16];
		        tmp16 += ((f16_temp - tmp16)*omega) + 
			                        ((c_3d_dev[16][0] - ux) * f16_temp ) / (rho2) * force;
			PDF_VAL(7) = tmp16;
			
			float_type tmp8 = shared_lbm2[share_x + 8];
		        tmp8 += ((f8_temp - tmp8)*omega) + 
			                        ((c_3d_dev[8][0] - ux) * f8_temp ) / (rho2) * force;
			PDF_VAL(17) = tmp8;

			float_type tmp17 = shared_lbm2[share_x + 17];
		        tmp17 += ((f17_temp - tmp17)*omega) + 
			                        ((c_3d_dev[17][0] - ux) * f17_temp ) / (rho2) * force;
			PDF_VAL(8) = tmp17;
			
			float_type tmp9 = shared_lbm2[share_x + 9];
		        tmp9 += ((f9_temp - tmp9)*omega) + 
			                        ((c_3d_dev[9][0] - ux) * f9_temp ) / (rho2) * force;
			PDF_VAL(18) = tmp9;

			float_type tmp18 = shared_lbm2[share_x + 18];
		        tmp18 += ((f18_temp - tmp18)*omega) + 
			                        ((c_3d_dev[18][0] - ux) * f18_temp ) / (rho2) * force;
			PDF_VAL(9) = tmp18;
		}
	}
}


/**
 * Kernel - Collide And Swap 2
 */
__global__ void	 collideAndSwapGPU_RegularBad(float_type omega, float_type force, cudaPitchedPtr solid_d, PDF pdf, size_t pitch, size_t ysize, int width, int height, int depth)
{
	/*int x = threadIdx.x;
	int grid_y = __mul24(blockIdx.y,gridDim.x);
	int z = (blockIdx.x + grid_y) / height;
	int y = (blockIdx.x + grid_y) - z * height;*/
	
        int x = threadIdx.x;
        int y = blockIdx.x;
        int z = blockIdx.y;
			
	float_type cu;	
	
	float_type omega_a = (1.f - omega);
			
	// Temporary memory for the particle distribution functions
	float_type f0_temp, f1_temp, f2_temp, f3_temp, f4_temp, f5_temp, f6_temp, f7_temp, f8_temp, f9_temp, f10_temp, f11_temp, f12_temp, f13_temp, f14_temp, f15_temp, f16_temp, f17_temp, f18_temp;
		     
	if (x < width && y < height && z < depth)
	{
		if(OBS_REF(int, solid_d, x, y, z) == 0)
		{
			int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, ysize);
	
			#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
			
			// Compute density from the particle distribution functions 
			float_type rho;
			getRho2(pdf, x, y, z, pitch, height, rho);
						 	
			// Compute velocity in x, y, and z directions from the particle distribution functions
			float_type ux; 
			getVelocityX2(pdf, x, y, z, pitch, height, rho, ux);
				
			float_type uy; 
			getVelocityY2(pdf, x, y, z, pitch, height, rho, uy);
			
			float_type uz;
			getVelocityZ2(pdf, x, y, z, pitch, height, rho, uz);
			
			float_type uSqr = ux*ux + uy*uy + uz*uz;
			
			float_type u_sqr = 1.5 * uSqr;
			
			float_type a = 4.5f;
				
			// Compute local equilibrium from rho and velocites 
					
			// 0
			cu = c_3d_dev[0][0]*ux + c_3d_dev[0][1]*uy + c_3d_dev[0][2]*uz;
			f0_temp = rho * t_3d_dev[0] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			
			// 1 
			cu = c_3d_dev[1][0]*ux + c_3d_dev[1][1]*uy + c_3d_dev[1][2]*uz;
			f1_temp = rho * t_3d_dev[1] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 2
			cu = c_3d_dev[2][0]*ux + c_3d_dev[2][1]*uy + c_3d_dev[2][2]*uz;
			f2_temp = rho * t_3d_dev[2] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			
			// 3
			cu = c_3d_dev[3][0]*ux + c_3d_dev[3][1]*uy + c_3d_dev[3][2]*uz;
			f3_temp = rho * t_3d_dev[3] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 4
			cu = c_3d_dev[4][0]*ux + c_3d_dev[4][1]*uy + c_3d_dev[4][2]*uz;
			f4_temp = rho * t_3d_dev[4] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 5
			cu = c_3d_dev[5][0]*ux + c_3d_dev[5][1]*uy + c_3d_dev[5][2]*uz;
			f5_temp = rho * t_3d_dev[5] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	

			// 6
			cu = c_3d_dev[6][0]*ux + c_3d_dev[6][1]*uy + c_3d_dev[6][2]*uz;
			f6_temp = rho * t_3d_dev[6] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);		
			
			// 7
			cu = c_3d_dev[7][0]*ux + c_3d_dev[7][1]*uy + c_3d_dev[7][2]*uz;
			f7_temp = rho * t_3d_dev[7] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 8 
			cu = c_3d_dev[8][0]*ux + c_3d_dev[8][1]*uy + c_3d_dev[8][2]*uz;
			f8_temp = rho * t_3d_dev[8] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 9
			cu = c_3d_dev[9][0]*ux + c_3d_dev[9][1]*uy + c_3d_dev[9][2]*uz;
			f9_temp = rho * t_3d_dev[9] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 10
			cu = c_3d_dev[10][0]*ux + c_3d_dev[10][1]*uy + c_3d_dev[10][2]*uz;
			f10_temp = rho * t_3d_dev[10] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 11
			cu = c_3d_dev[11][0]*ux + c_3d_dev[11][1]*uy + c_3d_dev[11][2]*uz;
			f11_temp = rho * t_3d_dev[11] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 12
			cu = c_3d_dev[12][0]*ux + c_3d_dev[12][1]*uy + c_3d_dev[12][2]*uz;
			f12_temp = rho * t_3d_dev[12] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	
			
			// 13
			cu = c_3d_dev[13][0]*ux + c_3d_dev[13][1]*uy + c_3d_dev[13][2]*uz;
			f13_temp = rho * t_3d_dev[13] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);

			// 14
			cu = c_3d_dev[14][0]*ux + c_3d_dev[14][1]*uy + c_3d_dev[14][2]*uz;
			f14_temp = rho * t_3d_dev[14] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	

			// 15
			cu = c_3d_dev[15][0]*ux + c_3d_dev[15][1]*uy + c_3d_dev[15][2]*uz;
			f15_temp = rho * t_3d_dev[15] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	

			// 16
			cu = c_3d_dev[16][0]*ux + c_3d_dev[16][1]*uy + c_3d_dev[16][2]*uz;
			f16_temp = rho * t_3d_dev[16] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);	

			// 17
			cu = c_3d_dev[17][0]*ux + c_3d_dev[17][1]*uy + c_3d_dev[17][2]*uz;
			f17_temp = rho * t_3d_dev[17] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			
			// 18
			cu = c_3d_dev[18][0]*ux + c_3d_dev[18][1]*uy + c_3d_dev[18][2]*uz;
			f18_temp = rho * t_3d_dev[18] * (1.f + (3.f*cu) + (a*cu*cu) - u_sqr);
			
			rho = rho * 0.333333333333f;
			
			// Relaxation of the particles distributions
			
			PDF_VAL(0) = (omega_a * PDF_VAL(0) + omega * f0_temp) + ( (c_3d_dev[0][0] - ux) * f0_temp ) / rho * force; ;

			float_type tmp1 = PDF_VAL(1); 
			float_type tmp10 = PDF_VAL(10); 
			
		        // Swap 1 and 10
		        PDF_VAL(1) = (omega_a * tmp10 + omega * f10_temp) + ( (c_3d_dev[10][0] - ux) * f10_temp ) / rho * force;
                        PDF_VAL(10) = (omega_a * tmp1 + omega * f1_temp) + ( (c_3d_dev[1][0] - ux) * f1_temp ) / rho * force;
            
		        float_type tmp2 = PDF_VAL(2); 
			float_type tmp11 = PDF_VAL(11); 
		    
		        // Swap 2 and 11
                        PDF_VAL(2) = (omega_a * tmp11 + omega * f11_temp) + ( (c_3d_dev[11][0] - ux) * f11_temp ) / rho * force;
                        PDF_VAL(11) = (omega_a * tmp2 + omega * f2_temp) + ( (c_3d_dev[2][0] - ux) * f2_temp ) / rho * force;
           
		        float_type tmp3 = PDF_VAL(3); 
			float_type tmp12 = PDF_VAL(12); 
			
			// Swap 3 and 12
                        PDF_VAL(3) = (omega_a * tmp12 + omega * f12_temp) + ( (c_3d_dev[12][0] - ux) * f12_temp ) / rho * force;
                        PDF_VAL(12) = (omega_a * tmp3 + omega * f3_temp) + ( (c_3d_dev[3][0] - ux) * f3_temp ) / rho * force;
           
			float_type tmp4 = PDF_VAL(4); 
			float_type tmp13 = PDF_VAL(13); 
			
			// Swap 4 and 13
                        PDF_VAL(4) = (omega_a * tmp13 + omega * f13_temp) + ( (c_3d_dev[13][0] - ux) * f13_temp ) / rho * force;
			PDF_VAL(13) = (omega_a * tmp4 + omega * f4_temp) + ( (c_3d_dev[4][0] - ux) * f4_temp ) / rho * force;
           
			float_type tmp5 = PDF_VAL(5); 
			float_type tmp14 = PDF_VAL(14); 
			 
			// Swap 5 and 14
                        PDF_VAL(5) = (omega_a * tmp14 + omega * f14_temp) + ( (c_3d_dev[14][0] - ux) * f14_temp ) / rho * force;
                        PDF_VAL(14) = (omega_a * tmp5 + omega * f5_temp) + ( (c_3d_dev[5][0] - ux) * f5_temp ) / rho * force;
           			
			float_type tmp6 = PDF_VAL(6); 
			float_type tmp15 = PDF_VAL(15); 
			
			// Swap 6 and 15
                        PDF_VAL(6) = (omega_a * tmp15 + omega * f15_temp) + ( (c_3d_dev[15][0] - ux) * f15_temp ) / rho * force;
                        PDF_VAL(15) = (omega_a * tmp6 + omega * f6_temp) + ( (c_3d_dev[6][0] - ux) * f6_temp ) / rho * force;
           
			float_type tmp7 = PDF_VAL(7); 
			float_type tmp16 = PDF_VAL(16); 
			
			// Swap 7 and 16
                        PDF_VAL(7) = (omega_a * tmp16 + omega * f16_temp) + ( (c_3d_dev[16][0] - ux) * f16_temp ) / rho * force;
                        PDF_VAL(16) = (omega_a * tmp7 + omega * f7_temp) + ( (c_3d_dev[7][0] - ux) * f7_temp ) / rho * force;
           
			float_type tmp8 = PDF_VAL(8); 
			float_type tmp17 = PDF_VAL(17); 
			
			// Swap 8 and 17
                        PDF_VAL(8) = (omega_a * tmp17 + omega * f17_temp) + ( (c_3d_dev[17][0] - ux) * f17_temp ) / rho * force;
                        PDF_VAL(17) = (omega_a * tmp8 + omega * f8_temp) + ( (c_3d_dev[8][0] - ux) * f8_temp ) / rho * force;
           
			float_type tmp9 = PDF_VAL(9); 
			float_type tmp18 = PDF_VAL(18); 
	
	                // Swap 9 and 18
                        PDF_VAL(9) = (omega_a * tmp18 + omega * f18_temp) + ( (c_3d_dev[18][0] - ux) * f18_temp ) / rho * force;
                        PDF_VAL(18) = (omega_a * tmp9 + omega * f9_temp) + ( (c_3d_dev[9][0] - ux) * f9_temp ) / rho * force;
		}
	}
}

/**
 * Kernel - Collide And Swap 5
 */
__global__ void	 collideAndSwapGPU_kernelPrecise2(float_type omega, float_type force, cudaPitchedPtr solid_d, PDF pdf, size_t pitch, size_t ysize, int width, int height, int depth)
{
	// Original blocking scheme
	/*int x = threadIdx.x;
	int z = (blockIdx.x + __mul24(blockIdx.y, gridDim.x)) / height;
	int y = (blockIdx.x + __mul24(blockIdx.y, gridDim.x)) - z * height;*/
        
        int x = threadIdx.x;
        int y = blockIdx.x;
        int z = blockIdx.y;
	
	if (x < width && y < height && z < depth) {
		if(OBS_REF(int, solid_d, x, y, z) == 0)
		{
			// Temporary memory for the particle distribution functions
			float_type f0_temp, f1_temp, f2_temp, f3_temp, f4_temp, f5_temp, f6_temp, f7_temp, f8_temp, 
				    f9_temp, f10_temp, f11_temp, f12_temp, f13_temp, f14_temp, f15_temp, f16_temp, f17_temp, f18_temp;
	        
			volatile int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, ysize);
	
			#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
			
			// Compute density from the particle distribution functions 
			
			float_type delta_rho;
			getRhoPrecise(pdf, x, y, z, pitch, height, delta_rho);
			
			float_type rho = delta_rho + 19.f;
										 	
			// Compute velocity in x, y, and z directions from the particle distribution functions
			
			float_type ux; 
			getVelocityXPrecise(pdf, x, y, z, pitch, height, rho, ux);
				
			float_type uy; 
			getVelocityYPrecise(pdf, x, y, z, pitch, height, rho, uy);
			
			float_type uz;
			getVelocityZPrecise(pdf, x, y, z, pitch, height, rho, uz);
			
			//float_type uSqr = ux*ux + uy*uy + uz*uz;
			//float_type u_sqr = 1.5f * uSqr;
			
			float_type u_sqr = (ux*ux + uy*uy + uz*uz) * 1.5f;
			

			// Compute local equilibrium from rho and velocites 
			
			float_type cu;	
					
			// 0
			f0_temp = t_3d_dev[0]*delta_rho - t_3d_dev[0]*(rho)*u_sqr;
		
			// 1 
			cu = c_3d_dev[1][0]*ux + c_3d_dev[1][1]*uy + c_3d_dev[1][2]*uz;			
			f1_temp = t_3d_dev[1]*delta_rho + t_3d_dev[1]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 2
			cu = c_3d_dev[2][0]*ux + c_3d_dev[2][1]*uy + c_3d_dev[2][2]*uz;
			f2_temp = t_3d_dev[2]*delta_rho + t_3d_dev[2]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 3
			cu = c_3d_dev[3][0]*ux + c_3d_dev[3][1]*uy + c_3d_dev[3][2]*uz;
			f3_temp = t_3d_dev[3]*delta_rho + t_3d_dev[3]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 4
			cu = c_3d_dev[4][0]*ux + c_3d_dev[4][1]*uy + c_3d_dev[4][2]*uz;
			f4_temp = t_3d_dev[4]*delta_rho + t_3d_dev[4]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);

			// 5
			cu = c_3d_dev[5][0]*ux + c_3d_dev[5][1]*uy + c_3d_dev[5][2]*uz;
			f5_temp = t_3d_dev[5]*delta_rho + t_3d_dev[5]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 6
			cu = c_3d_dev[6][0]*ux + c_3d_dev[6][1]*uy + c_3d_dev[6][2]*uz;
			f6_temp = t_3d_dev[6]*delta_rho + t_3d_dev[6]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
					
			// 7
			cu = c_3d_dev[7][0]*ux + c_3d_dev[7][1]*uy + c_3d_dev[7][2]*uz;
			f7_temp = t_3d_dev[7]*delta_rho + t_3d_dev[7]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 8 
			cu = c_3d_dev[8][0]*ux + c_3d_dev[8][1]*uy + c_3d_dev[8][2]*uz;
			f8_temp = t_3d_dev[8]*delta_rho + t_3d_dev[8]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);

			// 9
			cu = c_3d_dev[9][0]*ux + c_3d_dev[9][1]*uy + c_3d_dev[9][2]*uz;
			f9_temp = t_3d_dev[9]*delta_rho + t_3d_dev[9]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
				
			// 10
			cu = c_3d_dev[10][0]*ux + c_3d_dev[10][1]*uy + c_3d_dev[10][2]*uz;
			f10_temp = t_3d_dev[10]*delta_rho + t_3d_dev[10]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 11
			cu = c_3d_dev[11][0]*ux + c_3d_dev[11][1]*uy + c_3d_dev[11][2]*uz;
			f11_temp = t_3d_dev[11]*delta_rho + t_3d_dev[11]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 12
			cu = c_3d_dev[12][0]*ux + c_3d_dev[12][1]*uy + c_3d_dev[12][2]*uz;
			f12_temp = t_3d_dev[12]*delta_rho + t_3d_dev[12]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 13
			cu = c_3d_dev[13][0]*ux + c_3d_dev[13][1]*uy + c_3d_dev[13][2]*uz;
			f13_temp = t_3d_dev[13]*delta_rho + t_3d_dev[13]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);

			// 14
			cu = c_3d_dev[14][0]*ux + c_3d_dev[14][1]*uy + c_3d_dev[14][2]*uz;
			f14_temp = t_3d_dev[14]*delta_rho + t_3d_dev[14]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);

			// 15
			cu = c_3d_dev[15][0]*ux + c_3d_dev[15][1]*uy + c_3d_dev[15][2]*uz;
			f15_temp = t_3d_dev[15]*delta_rho + t_3d_dev[15]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 16
			cu = c_3d_dev[16][0]*ux + c_3d_dev[16][1]*uy + c_3d_dev[16][2]*uz;
			f16_temp = t_3d_dev[16]*delta_rho + t_3d_dev[16]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 17
			cu = c_3d_dev[17][0]*ux + c_3d_dev[17][1]*uy + c_3d_dev[17][2]*uz;
			f17_temp = t_3d_dev[17]*delta_rho + t_3d_dev[17]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			// 18
			cu = c_3d_dev[18][0]*ux + c_3d_dev[18][1]*uy + c_3d_dev[18][2]*uz;
			f18_temp = t_3d_dev[18]*delta_rho + t_3d_dev[18]*rho*((3.f*cu) + (4.5f*cu*cu) - u_sqr);
			
			cu = 3.0f * force / rho;
					
			// Relaxation of the particles distributions
			
			float_type tmp0 = PDF_VAL(0);
			PDF_VAL(0) = (tmp0 + omega * (f0_temp - tmp0)) + ((c_3d_dev[0][0] - ux) * (f0_temp + t_3d_dev[0]*19.f)) * cu;
			
			float_type tmp1 = PDF_VAL(1); 
			float_type tmp10 = PDF_VAL(10); 
			
		        // Swap 1 and 10
		        PDF_VAL(1) = (tmp10 + omega * (f10_temp - tmp10)) + ((c_3d_dev[10][0] - ux) * (f10_temp + t_3d_dev[10]*19.f)) * cu;
		        PDF_VAL(10) = (tmp1 + omega * (f1_temp - tmp1)) + ((c_3d_dev[1][0] - ux) * (f1_temp + t_3d_dev[1]*19.f)) * cu;
            
		        float_type tmp2 = PDF_VAL(2); 
			float_type tmp11 = PDF_VAL(11); 
		    
		        // Swap 2 and 11
                        PDF_VAL(2) = (tmp11 + omega * (f11_temp - tmp11)) + ((c_3d_dev[11][0] - ux) * (f11_temp + t_3d_dev[11]*19.f)) * cu;
                        PDF_VAL(11) = (tmp2 + omega * (f2_temp - tmp2)) + ((c_3d_dev[2][0] - ux) * (f2_temp + t_3d_dev[2]*19.f)) * cu;
           
		        float_type tmp3 = PDF_VAL(3); 
			float_type tmp12 = PDF_VAL(12); 
			
			// Swap 3 and 12
                        PDF_VAL(3) = (tmp12 + omega * (f12_temp - tmp12)) + ((c_3d_dev[12][0] - ux) * (f12_temp + t_3d_dev[12]*19.f)) * cu;
                        PDF_VAL(12) = (tmp3 + omega * (f3_temp - tmp3)) + ((c_3d_dev[3][0] - ux) * (f3_temp + t_3d_dev[3]*19.f)) * cu;
            
			float_type tmp4 = PDF_VAL(4); 
			float_type tmp13 = PDF_VAL(13); 
			
			// Swap 4 and 13
                        PDF_VAL(4) = (tmp13 + omega * (f13_temp - tmp13)) + ((c_3d_dev[13][0] - ux) * (f13_temp + t_3d_dev[13]*19.f)) * cu;
                        PDF_VAL(13) = (tmp4 + omega * (f4_temp - tmp4)) + ((c_3d_dev[4][0] - ux) * (f4_temp + t_3d_dev[4]*19.f)) * cu;
			
			float_type tmp5 = PDF_VAL(5); 
			float_type tmp14 = PDF_VAL(14); 
			 
			// Swap 5 and 14
                        PDF_VAL(5) = (tmp14 + omega * (f14_temp - tmp14)) + ((c_3d_dev[14][0] - ux) * (f14_temp + t_3d_dev[14]*19.f)) * cu;
                        PDF_VAL(14) = (tmp5 + omega * (f5_temp - tmp5)) + ((c_3d_dev[5][0] - ux) * (f5_temp + t_3d_dev[5]*19.f)) * cu;
           			
			float_type tmp6 = PDF_VAL(6); 
			float_type tmp15 = PDF_VAL(15); 
			
			// Swap 6 and 15
                        PDF_VAL(6) = (tmp15 + omega * (f15_temp - tmp15)) + ((c_3d_dev[15][0] - ux) * (f15_temp + t_3d_dev[15]*19.f)) * cu;
                        PDF_VAL(15) = (tmp6 + omega * (f6_temp - tmp6)) + ((c_3d_dev[6][0] - ux) * (f6_temp + t_3d_dev[6]*19.f)) * cu;
            
			float_type tmp7 = PDF_VAL(7); 
			float_type tmp16 = PDF_VAL(16); 
			
			// Swap 7 and 16
                        PDF_VAL(7) = (tmp16 + omega * (f16_temp - tmp16)) + ((c_3d_dev[16][0] - ux) * (f16_temp + t_3d_dev[16]*19.f)) * cu;
                        PDF_VAL(16) = (tmp7 + omega * (f7_temp - tmp7)) + ((c_3d_dev[7][0] - ux) * (f7_temp + t_3d_dev[7]*19.f)) * cu;
           
			float_type tmp8 = PDF_VAL(8); 
			float_type tmp17 = PDF_VAL(17); 
			
			// Swap 8 and 17
                        PDF_VAL(8) = (tmp17 + omega * (f17_temp - tmp17)) + ((c_3d_dev[17][0] - ux) * (f17_temp + t_3d_dev[17]*19.f)) * cu;
                        PDF_VAL(17) = (tmp8 + omega * (f8_temp - tmp8)) + ((c_3d_dev[8][0] - ux) * (f8_temp + t_3d_dev[8]*19.f)) * cu;
            
			float_type tmp9 = PDF_VAL(9); 
			float_type tmp18 = PDF_VAL(18); 
	
	                // Swap 9 and 18
                        PDF_VAL(9) = (tmp18 + omega * (f18_temp - tmp18)) + ((c_3d_dev[18][0] - ux) * (f18_temp + t_3d_dev[18]*19.f)) * cu;
                        PDF_VAL(18) = (tmp9 + omega * (f9_temp - tmp9)) + ((c_3d_dev[9][0] - ux) * (f9_temp + t_3d_dev[9]*19.f)) * cu;
		}
	}
}

/**
 * Kernel - Collide And Swap 5
 */
__global__ void	 collideAndSwapGPU_kernel5Double(double omega, double force, cudaPitchedPtr solid_d, PDF pdf, size_t pitch, size_t ysize, int width, int height, int depth)
{
	/*
	int x = threadIdx.x;
	//int grid_y = __mul24(blockIdx.y, gridDim.x);
	int grid_y = blockIdx.y * gridDim.x;
	int z = (blockIdx.x + grid_y) / height;
	int y = (blockIdx.x + grid_y) - z * height;
	*/
	int x = threadIdx.x;
    int y = blockIdx.x;
    int z = blockIdx.y;
	
	double swap_tmp;
			
	double cu;	
	
	double omega_a = (1.f - omega);
			
	// Temporary memory for the particle distribution functions
	double f0_temp, f1_temp, f2_temp, f3_temp, f4_temp, f5_temp, f6_temp, f7_temp, f8_temp, f9_temp, f10_temp, f11_temp, f12_temp, f13_temp, f14_temp, f15_temp, f16_temp, f17_temp, f18_temp;
	
	if (x < width && y < height && z < depth)
	{
		if(OBS_REF(int, solid_d, x, y, z) == 0)
		{
			int offset = (int) &PDF_REF(float_type, 0, x, y, z, pitch, ysize);
	
			#define PDF_VAL(n) (*((float_type*) ((char*)pdf.ptr[n] + offset)))
			
			// Compute density from the particle distribution functions 
			double delta_rho;
			getRho5Double(pdf, x, y, z, pitch, height, delta_rho);
			
			double rho = delta_rho + 19.;
										 	
			// Compute velocity in x, y, and z directions from the particle distribution functions
			double ux; 
			getVelocityX5Double(pdf, x, y, z, pitch, height, rho, ux);
				
			double uy; 
			getVelocityY5Double(pdf, x, y, z, pitch, height, rho, uy);
			
			double uz;
			getVelocityZ5Double(pdf, x, y, z, pitch, height, rho, uz);
			
			double uSqr = ux*ux + uy*uy + uz*uz;
			
			double u_sqr = 1.5 * uSqr;
			
			double a = 4.5;
				
			// Compute local equilibrium from rho and velocites 
					
			// 0
			f0_temp = t_3d_dev[0]*delta_rho - t_3d_dev[0]*(rho)*u_sqr;
		
			// 1 
			cu = c_3d_dev[1][0]*ux + c_3d_dev[1][1]*uy + c_3d_dev[1][2]*uz;			
			f1_temp = t_3d_dev[1]*delta_rho + t_3d_dev[1]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 2
			cu = c_3d_dev[2][0]*ux + c_3d_dev[2][1]*uy + c_3d_dev[2][2]*uz;
			f2_temp = t_3d_dev[2]*delta_rho + t_3d_dev[2]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 3
			cu = c_3d_dev[3][0]*ux + c_3d_dev[3][1]*uy + c_3d_dev[3][2]*uz;
			f3_temp = t_3d_dev[3]*delta_rho + t_3d_dev[3]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 4
			cu = c_3d_dev[4][0]*ux + c_3d_dev[4][1]*uy + c_3d_dev[4][2]*uz;
			f4_temp = t_3d_dev[4]*delta_rho + t_3d_dev[4]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);

			// 5
			cu = c_3d_dev[5][0]*ux + c_3d_dev[5][1]*uy + c_3d_dev[5][2]*uz;
			f5_temp = t_3d_dev[5]*delta_rho + t_3d_dev[5]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 6
			cu = c_3d_dev[6][0]*ux + c_3d_dev[6][1]*uy + c_3d_dev[6][2]*uz;
			f6_temp = t_3d_dev[6]*delta_rho + t_3d_dev[6]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
					
			// 7
			cu = c_3d_dev[7][0]*ux + c_3d_dev[7][1]*uy + c_3d_dev[7][2]*uz;
			f7_temp = t_3d_dev[7]*delta_rho + t_3d_dev[7]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 8 
			cu = c_3d_dev[8][0]*ux + c_3d_dev[8][1]*uy + c_3d_dev[8][2]*uz;
			f8_temp = t_3d_dev[8]*delta_rho + t_3d_dev[8]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);

			// 9
			cu = c_3d_dev[9][0]*ux + c_3d_dev[9][1]*uy + c_3d_dev[9][2]*uz;
			f9_temp = t_3d_dev[9]*delta_rho + t_3d_dev[9]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
				
			// 10
			cu = c_3d_dev[10][0]*ux + c_3d_dev[10][1]*uy + c_3d_dev[10][2]*uz;
			f10_temp = t_3d_dev[10]*delta_rho + t_3d_dev[10]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 11
			cu = c_3d_dev[11][0]*ux + c_3d_dev[11][1]*uy + c_3d_dev[11][2]*uz;
			f11_temp = t_3d_dev[11]*delta_rho + t_3d_dev[11]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 12
			cu = c_3d_dev[12][0]*ux + c_3d_dev[12][1]*uy + c_3d_dev[12][2]*uz;
			f12_temp = t_3d_dev[12]*delta_rho + t_3d_dev[12]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 13
			cu = c_3d_dev[13][0]*ux + c_3d_dev[13][1]*uy + c_3d_dev[13][2]*uz;
			f13_temp = t_3d_dev[13]*delta_rho + t_3d_dev[13]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);

			// 14
			cu = c_3d_dev[14][0]*ux + c_3d_dev[14][1]*uy + c_3d_dev[14][2]*uz;
			f14_temp = t_3d_dev[14]*delta_rho + t_3d_dev[14]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);

			// 15
			cu = c_3d_dev[15][0]*ux + c_3d_dev[15][1]*uy + c_3d_dev[15][2]*uz;
			f15_temp = t_3d_dev[15]*delta_rho + t_3d_dev[15]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 16
			cu = c_3d_dev[16][0]*ux + c_3d_dev[16][1]*uy + c_3d_dev[16][2]*uz;
			f16_temp = t_3d_dev[16]*delta_rho + t_3d_dev[16]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 17
			cu = c_3d_dev[17][0]*ux + c_3d_dev[17][1]*uy + c_3d_dev[17][2]*uz;
			f17_temp = t_3d_dev[17]*delta_rho + t_3d_dev[17]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			// 18
			cu = c_3d_dev[18][0]*ux + c_3d_dev[18][1]*uy + c_3d_dev[18][2]*uz;
			f18_temp = t_3d_dev[18]*delta_rho + t_3d_dev[18]*(rho)*((3.*cu) + (a*cu*cu) - u_sqr);
			
			rho = rho * 0.333333333333333333333333333333;
			
			// Relaxation of the particles distributions
			
			double tmp0 = PDF_VAL(0);
			PDF_VAL(0) = (tmp0 + omega * (f0_temp - tmp0)) + ((c_3d_dev[0][0] - ux) * (f0_temp + t_3d_dev[0]*19.)) / rho * force;
		
			double tmp1 = PDF_VAL(1); 
			double tmp10 = PDF_VAL(10); 
			
	                // Swap 1 and 10
		        PDF_VAL(1) = (tmp10 + omega * (f10_temp - tmp10)) + ((c_3d_dev[10][0] - ux) * (f10_temp + t_3d_dev[10]*19.)) / rho * force;
		        PDF_VAL(10) = (tmp1 + omega * (f1_temp - tmp1)) + ((c_3d_dev[1][0] - ux) * (f1_temp + t_3d_dev[1]*19.)) / rho * force;
            
		        double tmp2 = PDF_VAL(2); 
			double tmp11 = PDF_VAL(11); 
		    
		        // Swap 2 and 11
                        PDF_VAL(2) = (tmp11 + omega * (f11_temp - tmp11)) + ((c_3d_dev[11][0] - ux) * (f11_temp + t_3d_dev[11]*19.)) / rho * force;
                        PDF_VAL(11) = (tmp2 + omega * (f2_temp - tmp2)) + ((c_3d_dev[2][0] - ux) * (f2_temp + t_3d_dev[2]*19.)) / rho * force;
            
		        double tmp3 = PDF_VAL(3); 
			double tmp12 = PDF_VAL(12); 
			
			// Swap 3 and 12
                        PDF_VAL(3) = (tmp12 + omega * (f12_temp - tmp12)) + ((c_3d_dev[12][0] - ux) * (f12_temp + t_3d_dev[12]*19.)) / rho * force;
                        PDF_VAL(12) = (tmp3 + omega * (f3_temp - tmp3)) + ((c_3d_dev[3][0] - ux) * (f3_temp + t_3d_dev[3]*19.)) / rho * force;
            
			double tmp4 = PDF_VAL(4); 
			double tmp13 = PDF_VAL(13); 
			
			// Swap 4 and 13
                        PDF_VAL(4) = (tmp13 + omega * (f13_temp - tmp13)) + ((c_3d_dev[13][0] - ux) * (f13_temp + t_3d_dev[13]*19.)) / rho * force;
                        PDF_VAL(13) = (tmp4 + omega * (f4_temp - tmp4)) + ((c_3d_dev[4][0] - ux) * (f4_temp + t_3d_dev[4]*19.)) / rho * force;
			
			double tmp5 = PDF_VAL(5); 
			double tmp14 = PDF_VAL(14); 
			 
			// Swap 5 and 14
                        PDF_VAL(5) = (tmp14 + omega * (f14_temp - tmp14)) + ((c_3d_dev[14][0] - ux) * (f14_temp + t_3d_dev[14]*19.)) / rho * force;
                        PDF_VAL(14) = (tmp5 + omega * (f5_temp - tmp5)) + ((c_3d_dev[5][0] - ux) * (f5_temp + t_3d_dev[5]*19.)) / rho * force;
            		
			double tmp6 = PDF_VAL(6); 
			double tmp15 = PDF_VAL(15); 
			
			// Swap 6 and 15
                        PDF_VAL(6) = (tmp15 + omega * (f15_temp - tmp15)) + ((c_3d_dev[15][0] - ux) * (f15_temp + t_3d_dev[15]*19.)) / rho * force;
                        PDF_VAL(15) = (tmp6 + omega * (f6_temp - tmp6)) + ((c_3d_dev[6][0] - ux) * (f6_temp + t_3d_dev[6]*19.)) / rho * force;
            
			double tmp7 = PDF_VAL(7); 
			double tmp16 = PDF_VAL(16); 
			
			// Swap 7 and 16
                        PDF_VAL(7) = (tmp16 + omega * (f16_temp - tmp16)) + ((c_3d_dev[16][0] - ux) * (f16_temp + t_3d_dev[16]*19.)) / rho * force;
                        PDF_VAL(16) = (tmp7 + omega * (f7_temp - tmp7)) + ((c_3d_dev[7][0] - ux) * (f7_temp + t_3d_dev[7]*19.)) / rho * force;
            
			double tmp8 = PDF_VAL(8); 
			double tmp17 = PDF_VAL(17); 
			
			// Swap 8 and 17
                        PDF_VAL(8) = (tmp17 + omega * (f17_temp - tmp17)) + ((c_3d_dev[17][0] - ux) * (f17_temp + t_3d_dev[17]*19.)) / rho * force;
                        PDF_VAL(17) = (tmp8 + omega * (f8_temp - tmp8)) + ((c_3d_dev[8][0] - ux) * (f8_temp + t_3d_dev[8]*19.)) / rho * force;
            
			double tmp9 = PDF_VAL(9); 
			double tmp18 = PDF_VAL(18); 
	
	                // Swap 9 and 18
                        PDF_VAL(9) = (tmp18 + omega * (f18_temp - tmp18)) + ((c_3d_dev[18][0] - ux) * (f18_temp + t_3d_dev[18]*19.)) / rho * force;
                        PDF_VAL(18) = (tmp9 + omega * (f9_temp - tmp9)) + ((c_3d_dev[9][0] - ux) * (f9_temp + t_3d_dev[9]*19.)) / rho * force;			
		}
	}
}

#endif
