#ifndef __CLOTHOID_H__
#define __CLOTHOID_H__

#include <vector>

#include "vmath.h"

namespace clothoid
{
    struct connector_t
    {
        Vec2d T, //tangent vector
              p; //position

        connector_t(Vec2d _p, Vec2d _T)
        {
            p = _p; 
            T = _T;
        }

        connector_t()
        {
        }
    };

    struct clothoid_point_t
    {
        double t, curvature;
        Vec2d pos;
        };

    class ClothoidPair
    {
//        private:
        public:
            double length_total, 
                   a0, a1, //Scaling factors
                   alpha0, alpha1, //Angle of starting tangent
                   t0, t1, //Curve limits
                   g_diff; //Difference between the actual g, and the limited g (g_diff = length of beginning straight line segment)

            connector_t p0, p1;  // Point at the end of "g" and "h" respectively, with normal
            bool flip0, reverse; //reverse indicates that the clothoids in the pair is switched

        private:
            void construct(const connector_t& pa, const connector_t& pb, const Vec2d& ctrl, double tau);

        public:
            ClothoidPair(const Vec2d& pa, const Vec2d& pb, const Vec2d& Ta, const Vec2d& Tb, const Vec2d& ctrl, double tau = 0.75);
            ClothoidPair(const connector_t& pa, const connector_t& pb, const Vec2d& ctrl, double tau = 0.75);
            clothoid_point_t lookup(double t); //Lookup a coordinate in the clothoid pair
            double length();
            double integratedCurvature();
    };

    class ClothoidSpline
    {
        private:
            std::vector<connector_t> connectors;
            std::vector<Vec2d> controlPoints;
            std::vector<double> lengths;          // Accumulated curve length at each clothoid pair's starting point
            std::vector<ClothoidPair> clothoidPairs;

        private:
            void construct(const std::vector<Vec2d>& controlPoints, double tau);

        public:
            ClothoidSpline(const std::vector<Vec2d>& controlPoints, double tau = 0.75);

            clothoid_point_t lookup(double t);
            size_t lookupClothoidPairIndex(double t);
            double length();
    };
}
#endif
