#include "conf.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>

#define LCASE \
         'a': \
    case 'b': \
    case 'c': \
    case 'd': \
    case 'e': \
    case 'f': \
    case 'g': \
    case 'h': \
    case 'i': \
    case 'j': \
    case 'k': \
    case 'l': \
    case 'm': \
    case 'n': \
    case 'o': \
    case 'p': \
    case 'q': \
    case 'r': \
    case 's': \
    case 't': \
    case 'u': \
    case 'v': \
    case 'w': \
    case 'x': \
    case 'y': \
    case 'z'

#define UCASE \
         'A': \
    case 'B': \
    case 'C': \
    case 'D': \
    case 'E': \
    case 'F': \
    case 'G': \
    case 'H': \
    case 'I': \
    case 'J': \
    case 'K': \
    case 'L': \
    case 'M': \
    case 'N': \
    case 'O': \
    case 'P': \
    case 'Q': \
    case 'R': \
    case 'S': \
    case 'T': \
    case 'U': \
    case 'V': \
    case 'W': \
    case 'X': \
    case 'Y': \
    case 'Z'

#define DIGIT \
         '0': \
    case '1': \
    case '2': \
    case '3': \
    case '4': \
    case '5': \
    case '6': \
    case '7': \
    case '8': \
    case '9'

#define SPECIAL \
         '.': \
    case ',': \
    case '\'': \
    case '`': \
    case '[': \
    case ']': \
    case '(': \
    case ')': \
    case '<': \
    case '>': \
    case '_': \
    case '+': \
    case '-': \
    case '*': \
    case '/': \
    case '=': \
    case '|': \
    case '?': \
    case '!': \
    case '&': \
    case '%': \
    case '$': \
    case '@': \
    case '~': \
    case '^'

#define SPACE \
         '\r': \
    case '\t': \
    case ' '

#define WORD \
         LCASE: \
    case UCASE: \
    case DIGIT: \
    case SPECIAL

/* mark the beginning of a string */
#define MARK p = &text[pos-1]

/* terminate name string */
#define NTERM text[pos-1] = '\0'; n->nam = p

/* terminate value string and eat trailing spaces */
#define VTERM \
    VQTERM; j = pos-2; \
    while(1) { \
        switch(text[j]) { case '\n': case SPACE: text[j--] = '\0'; continue; } \
        break; \
    }

/* don't include quote char */
#define QMARK p = &text[pos];

/* 0-terminate at quote char, make value NULL if length zero */
#define VQTERM text[pos-1] = '\0'; if(!p[0]) p = NULL; n->val = p

/* names can't be empty */
#define NQTERM text[pos-1] = '\0'; if(!p[0]) ERR("empty name"); n->nam = p

/* allocate a new node and initialize it,
also updating child and next references of previous nodes.
allocate space for a char pointer before the first Conf,
which will hold a reference to the text(for deallocation) */
#define NEW \
    if(num_nodes >= max_nodes) { \
        if(max_nodes)  max_nodes *= 2; \
        else max_nodes = 64; \
        data = (char *)realloc(data, sizeof(char *)+sizeof(Conf)*max_nodes); \
        nodes = (Conf *)(data + sizeof(char *)); \
    } \
    n = &nodes[num_nodes]; \
    n->nam = n->val = NULL; \
    n->u.idx.next = n->u.idx.child = -1; \
    if(prev[i]) prev[i]->u.idx.next = num_nodes; \
    else if(i) prev[i-1]->u.idx.child = num_nodes; \
    prev[i] = n; \
    ++num_nodes; \
    aok = 1

/* descend into new block, 'prev' node here yet to be encountered */
#define DOWN prev[++i] = NULL

#define UP --i; aok = 1

#define ERR(what) do { fprintf(stderr, "line %d: %s\n", line, what); goto err; } while(0)
#define ERR_CHAR(what) do { fprintf(stderr, "line %d: unexpected char '%c' while %s\n", line, text[pos-1], what); goto err; } while(0)
#define ERR_EOF(what) do { fprintf(stderr, "line %d: unexpected eof while %s\n", line, what); goto err; } while(0)

enum {
    S_expect_name,
    S_expect_name_or_brace,
    S_expect_value_or_newline_or_brace,
    S_expect_value,
    S_read_name,
    S_read_quoted_name,
    S_read_value,
    S_read_quoted_value
};


Conf *Conf::parse(char *text) {
    int i, j, pos, line, next, aok;
    char *p = NULL;
    
    Conf *n = NULL, *nodes = NULL;
    Conf *prev[32]; /* stack of previous nodes */
    int num_nodes = 0;
    int max_nodes = 0;
    char *data = NULL;

    i = 0;          /* an index to the prev stack, modified as depth changes */
    prev[0] = NULL; /* no nodes encountered yet at this level */
    pos = 0;        /* indexing the text array */
    line = 1;       /* keep track of line numbers for error reporting */
    next = S_expect_name; /* set initial state */
    aok = 0;        /* is a semicolon ok now? NEW sets this to 1 */
    
resume:
    switch(next) {
        case S_expect_name:
expect_name:
            switch(text[pos++]) {
                case '#': next = S_expect_name; goto comment;
                case '\n': ++line;
                case SPACE: goto expect_name;
                case WORD: MARK; NEW; goto read_name;
                case '"': QMARK; NEW; goto read_quoted_name;
                case '}': if(!i) ERR_CHAR("expecting name"); UP; goto expect_name;
                case '\0': if(i) ERR_EOF("expecting name"); goto end;
                case ';': if(!prev[i] || !aok) ERR_CHAR("expecting name"); aok = 0; goto expect_name;
                default: ERR_CHAR("expecting name");
            }
        case S_expect_name_or_brace:
expect_name_or_brace:
            switch(text[pos++]) {
                case '#': next = S_expect_name_or_brace; goto comment;
                case '\n': ++line;
                case SPACE: goto expect_name_or_brace;
                case '{': DOWN; goto expect_name;
                case '}': if(!i) ERR_CHAR("expecting name or '{'"); UP; goto expect_name;
                case WORD: MARK; NEW; goto read_name;
                case '"': QMARK; NEW; goto read_quoted_name;
                case '\0': if(i) ERR_EOF("expecting name or '{'"); goto end;
                case ';': if(!aok) ERR_CHAR("expecting name"); aok = 0; goto expect_name_or_brace;
                default: ERR_CHAR("expecting name or '{'");
            }
        case S_expect_value_or_newline_or_brace:
expect_value_or_newline_or_brace:
            switch(text[pos++]) {
                case '#': next = S_expect_name_or_brace; goto comment;
                case '\n': ++line; goto expect_name_or_brace;
                case '{': DOWN; goto expect_name;
                case '}': if(!i) ERR_CHAR("reading name"); UP; goto expect_name;
                case ';': aok = 0; goto expect_name;
                case '"': QMARK; goto read_quoted_value;
                case WORD: MARK; goto read_value;
                case ':': goto expect_value;
                case SPACE: goto expect_value_or_newline_or_brace;
                default: ERR_CHAR("expecting value, newline or '{'");
            }
        case S_expect_value:
expect_value:
            switch(text[pos++]) {
                case '#': next = S_expect_value; goto comment;
                case '\n': ++line;
                case SPACE: goto expect_value;
                case WORD: MARK; goto read_value;
                case '"': QMARK; goto read_quoted_value;
                default: ERR_CHAR("expecting value");
            }
        case S_read_name:
read_name:
            switch(text[pos++]) {
                case '#': NTERM; next = S_expect_name_or_brace; goto comment;
                case '\n': ++line; NTERM; goto expect_name_or_brace;
                case ';': NTERM; aok = 0; goto expect_name;
                case '{': NTERM; DOWN; goto expect_name;
                case '}': if(!i) ERR_CHAR("reading name"); NTERM; UP; goto expect_name;
                case '"': NTERM; QMARK; goto read_quoted_value;
                case ':': NTERM; goto expect_value;
                case SPACE: NTERM; goto expect_value_or_newline_or_brace;
                case WORD: goto read_name;
                default: ERR_CHAR("reading name");
            }
        case S_read_quoted_name:
read_quoted_name:
            switch(text[pos++]) {
                case '"': NQTERM; goto expect_value_or_newline_or_brace;
                case '\0': ERR_EOF("expecting '\"'"); goto end;
                case '\n': ++line;
                default: goto read_quoted_name;
            }
        case S_read_value:
read_value:
            switch(text[pos++]) {
                case '#': VTERM; next = S_expect_name_or_brace; goto comment;
                case '\n': ++line; VTERM; goto expect_name_or_brace;
                case ';': VTERM; aok = 0; goto expect_name;
                case '{': VTERM; DOWN; goto expect_name;
                case '}': if(!i) ERR_CHAR("reading value"); VTERM; UP; goto expect_name;
                case SPACE:
                case WORD: goto read_value;
                default: ERR_CHAR("reading value");
            }
        case S_read_quoted_value:
read_quoted_value:
            switch(text[pos++]) {
                case '"': VQTERM; goto expect_name_or_brace;
                case '\0': ERR_EOF("expecting '\"'"); goto end;
                case '\n': ++line;
                default: goto read_quoted_value;
            }
    }
comment:
    switch(text[pos++]) {
        case '\n': ++line; goto resume;
        case '\0': if(i) ERR_EOF("reading comment"); goto end;
        default: goto comment;
    }
    
err:
    if(data)
		::free(data);
    return NULL;
    
end:
    for(i = 0; i < num_nodes; ++i) {
        n = nodes+i;
        
        pos = n->u.idx.child;
        prev[0] = pos >= 0 ? nodes+pos : NULL;

        pos = n->u.idx.next;
        prev[1] = pos >= 0 ? nodes+pos : NULL;
        
        n->u.ptr.child = prev[0];
        n->u.ptr.next = prev[1];
    }
    
    /*printf("num_nodes: %d\n", num_nodes);*/
    *(char **)data = NULL;
    return nodes;
}

static char *loadfile(const char *file) {
    FILE *fp;
    char *text;
    long size;
    
    fp = fopen(file, "rb");
    if(!fp) {
        fprintf(stderr, "error: unable to open file '%s'\n", file);
        return NULL;
    }
    fseek(fp, 0, SEEK_END);
    size = ftell(fp);
    rewind(fp);
    text = (char *)malloc(size+1);
    fread(text, 1, size, fp);
    text[size] = '\0';
    fclose(fp);
    
    return text;
}

Conf *Conf::load(const char *file) {
    char *text = loadfile(file);
    if(text) {
        Conf *conf = parse(text);
        if(conf) {
            char *data = (char *)conf - sizeof(char *);
            *(char **)data = text;
            return conf;
        }
        fprintf(stderr, "error encountered while parsing '%s'\n", file);
		::free(text);
    }
    return NULL;
}

void Conf::free(Conf *conf) {
    char *data = (char *)conf - sizeof(char *);
    char *text = *(char **)data;
	if(text) ::free(text);
	::free(data);
}
