#ifndef NV_NPP_UTIL_STOP_WATCH_BASE_H
#define NV_NPP_UTIL_STOP_WATCH_BASE_H
/*
* Copyright 2008-2009 NVIDIA Corporation.  All rights reserved.
*
* NOTICE TO USER:
*
* This source code is subject to NVIDIA ownership rights under U.S. and
* international Copyright laws.
*
* NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
* CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
* IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
* REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
* MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
* IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
* OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
* OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
* OR PERFORMANCE OF THIS SOURCE CODE.
*
* U.S. Government End Users.  This source code is a "commercial item" as
* that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
* "commercial computer software" and "commercial computer software
* documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
* and is provided to the U.S. Government only as a commercial end item.
* Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
* 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
* source code with only those rights set forth herein.
*/


namespace npp
{

    /// Simple stop watch class
    ///     This class uses high-precision timers. It is implemented
    /// using templates and inline functions to cause minimal call-overhead
    /// and provide the most accurate timings.
    template<class OSPolicy>
    class StopWatchBase : public OSPolicy 
    {
    public:

        // generic, specialized type
        typedef StopWatchBase<OSPolicy>   SelfType;
        // generic, specialized type pointer
        typedef StopWatchBase<OSPolicy>*  SelfTypePtr;

    public:

        //! Constructor, default
        StopWatchBase();

        // Destructor
        ~StopWatchBase();

    public:

        //! Start time measurement
        inline void start();

        //! Stop time measurement
        inline void stop();

        //! Reset time counters to zero
        inline void reset();

        //! Time in msec. after start. If the stop watch is still running (i.e. there
        //! was no call to stop()) then the elapsed time is returned, otherwise the
        //! time between the last start() and stop call is returned
        inline const double elapsed() const;

    private:

        //! Constructor, copy (not implemented)
        StopWatchBase( const StopWatchBase&);

        //! Assignment operator (not implemented)
        StopWatchBase& operator=( const StopWatchBase&);
    };

// include, implementation
#include "StopWatchBase.inl"

} // npp namespace 

#endif // NV_NPP_UTIL_STOP_WATCH_BASE_H

