#version 410 core

uniform float dim;
uniform mat4 modelviewMatrix2;
uniform mat4 projMatrix2;
uniform sampler2D normalmap;
uniform sampler2D snowmap;
uniform sampler2D groundmap;
uniform sampler2D terrain;

#ifdef _VERTEX_
layout(location = 0) in vec3 in_Position;
out float alpha;
out vec2 vNorm;
out vec2 base;
void main() {
    vec4 vertex;
    vec2 norm_coord, base_coord;
    
    float height = texture(terrain, vertex.xy/256).x;
    
    vertex = vec4(in_Position, 1.0);
    float w = vertex.w;
    vertex.w = 1.0;
    if(vertex.x > 100 && vertex.z > 100)
        vertex.y+=100;
    
    alpha = w*10.0;
    if(w > 0.1) {
       vertex.y += (w-0.1)*1.5;
       alpha = 1.0;
    }
    
    norm_coord = vertex.xz * (1.0 / dim);
    base_coord = vertex.xz * (1.0 / 16.0);

    
	vNorm = norm_coord;
    base = base_coord;
	gl_Position = projMatrix2*modelviewMatrix2 *vertex; 
}
#endif

#ifdef _FRAGMENT_
in float alpha;
in vec2 vNorm;
in vec2 base;
out vec4 fragColor;

void main() {
    vec3 norm, lightdir;
    vec4 snowcolor, groundcolor;
    
    norm = vec3(texture(normalmap, vNorm));
    snowcolor = texture(snowmap, base);
    groundcolor = texture(groundmap, base);
    
    vec4 c = groundcolor*(1.0-alpha) + snowcolor*alpha;
    c.w = 1.0;

    lightdir = normalize(vec3(1.0, 0.2, 0.0));
    c *= max(dot(norm, lightdir), 0.0)*1.2;

    
	fragColor = c;
	//fragColor = vec4(0.5, 0.5, 0.5, 1.0);
}

#endif
